# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from copy import deepcopy
import pytest
import shutil

from hgitaly.oid import (
    blob_oid,
    tree_oid,
)
from hgitaly.revision import (
    gitlab_revision_changeset,
    ZERO_SHA,
    ZERO_SHA_STR,
)
from hgitaly.stream import WRITE_BUFFER_SIZE

from hgitaly.stub.shared_pb2 import (
    PaginationParameter,
    Repository,
)
from hgitaly.stub.blob_pb2 import (
    GetBlobRequest,
    GetBlobsRequest,
)
from hgitaly.stub.commit_pb2 import (
    GetTreeEntriesRequest,
    TreeEntryRequest,
)
from hgitaly.stub.analysis_pb2 import (
    CheckBlobsGeneratedRequest,
)
from hgitaly.stub.analysis_pb2_grpc import AnalysisServiceStub
from hgitaly.stub.blob_pb2_grpc import BlobServiceStub
from hgitaly.stub.commit_pb2_grpc import CommitServiceStub

from . import skip_comparison_tests
if skip_comparison_tests():  # pragma no cover
    pytestmark = pytest.mark.skip

parametrize = pytest.mark.parametrize


def concat_data_fields(responses):
    """Callback for RpcHelper.assert_compare_aggregated.

    Works for all response gRPC messages with a `data` field.
    """
    return b''.join(r.data for r in responses)


def remove_data_field(response):
    """Callback for RpcHelper.assert_compare_aggregated.

    Allows to compare all fields but `data`.
    """
    response.data = b''


def oid_mapping(gitaly_comp, rev_paths):
    """Provide the OID mappings between Mercurial and Git for blobs.

    :param fixture: an instance of `GitalyComparison`
    :param rev_paths: an iterable of tuples (revision, path, type), where
      type is 'tree' or 'blob'
    :return: a :class:`dict` mapping Mercurial oids to Git oids

    Does not need HGitaly blob/tree methods to work, only those of Gitaly.
    """
    res = {}
    gitaly_meth = CommitServiceStub(gitaly_comp.gitaly_channel).TreeEntry
    hg_repo = gitaly_comp.hg_repo_wrapper.repo
    hg_oid_meths = dict(tree=tree_oid, blob=blob_oid)
    for rev, path, t in rev_paths:
        git_rev = gitaly_comp.hg_git.map_git_get(rev)
        if git_rev is None:  # happens for symbolic revisions (branches etc).
            git_rev = rev
        changeset = gitlab_revision_changeset(hg_repo, rev)
        hg_oid = hg_oid_meths[t](hg_repo, changeset.hex().decode(), path)
        gitaly_resp = gitaly_meth(
            TreeEntryRequest(repository=gitaly_comp.gitaly_repo,
                             revision=git_rev,
                             path=path,
                             limit=1)
        )
        res[hg_oid] = next(iter(gitaly_resp)).oid
    return res


def oid_normalizer(oid2git):
    """Return a response normalizer for oid fields.

    :param oid_mapping: :class:`dict` instance mapping HGitaly OIDs to
      Gitaly OIDs
    """
    def normalizer(rpc_helper, responses, vcs='hg'):
        if vcs != 'hg':
            return

        for entry in responses:
            if entry.oid:
                entry.oid = oid2git[entry.oid]

    return normalizer


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_tree_entry_request(gitaly_rhgitaly_comparison, hg_server):
    fixture = gitaly_rhgitaly_comparison

    wrapper = fixture.hg_repo_wrapper
    wrapper.write_commit('foo', message="Some foo")
    sub = (wrapper.path / 'sub')
    sub.mkdir()
    (sub / 'bar').write_text('bar content')
    (sub / 'ba2').write_text('ba2 content')
    # TODO OS indep for paths (actually TODO make wrapper.commit easier to
    # use, e.g., check how to make it accept patterns)
    wrapper.commit(rel_paths=['sub/bar', 'sub/ba2'],
                   message="zebar", add_remove=True)

    default_rev = b'branch/default'
    oid2git = oid_mapping(fixture,
                          ((default_rev, b'foo', 'blob'),
                           (default_rev, b'sub', 'tree'),
                           (default_rev, b'sub/bar', 'blob'),
                           (default_rev, b'sub/ba2', 'blob'),
                           ))
    normalizer = oid_normalizer(oid2git)

    rpc_helper = fixture.rpc_helper(stub_cls=CommitServiceStub,
                                    hg_server=hg_server,
                                    method_name='TreeEntry',
                                    request_cls=TreeEntryRequest,
                                    request_defaults=dict(
                                        revision=default_rev,
                                        limit=0,
                                        max_size=0),
                                    streaming=True,
                                    normalizer=normalizer,
                                    chunked_fields_remover=remove_data_field,
                                    chunks_concatenator=concat_data_fields,
                                    )
    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    # precondition for the test: mirror worked
    assert fixture.git_repo.branch_titles() == {default_rev: b"zebar"}

    # basic case
    for path in (b'sub', b'sub/bar', b'sub/'):
        assert_compare(path=path)

    for path in (b'.', b'do-not-exist'):
        assert_compare_errors(path=path)

    # limit and max_size (does not apply to Trees)
    assert_compare(path=b'foo', limit=4)
    assert_compare_errors(path=b'foo', max_size=4)
    assert_compare(path=b'sub', max_size=1)

    # unknown revision
    assert_compare_errors(path=b'sub', revision=b'unknown')
    assert_compare_errors(path=b'sub', revision=ZERO_SHA)

    # chunking for big Blob entry
    wrapper.write_commit('bigfile', message="A big file with 2 or 3 chunks",
                         content=b'a' * 1023 + b'ff' * 65536)
    # default_rev now resolves to another changeset, hence a different
    # HGitaly OID given the current implementation
    oid2git.update(oid_mapping(fixture, ((default_rev, b'bigfile', 'blob'),
                                         (default_rev, b'sub', 'tree'),
                                         (default_rev, b'sub/bar', 'blob'),
                                         (default_rev, b'sub/ba2', 'blob'),
                                         (default_rev, b'foo', 'blob'),
                                         )))
    rpc_helper.assert_compare_aggregated(path=b'bigfile',
                                         compare_first_chunks=False)

    #
    # reusing content to test GetTreeEntries, hence with a new rpc_helper
    #

    def gte_normalizer(rpc_helper, responses, vcs='hg'):
        for resp in responses:
            for entry in resp.entries:
                if entry.oid and vcs == 'hg':
                    entry.oid = oid2git[entry.oid]

    rpc_helper = fixture.rpc_helper(stub_cls=CommitServiceStub,
                                    hg_server=hg_server,
                                    method_name='GetTreeEntries',
                                    request_cls=GetTreeEntriesRequest,
                                    request_defaults=dict(
                                        revision=default_rev,
                                        ),
                                    streaming=True,
                                    normalizer=gte_normalizer,
                                    )
    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors

    for path in (b'.', b'sub'):
        for recursive in (False, True):
            for skip_flat in (False, True):
                assert_compare(path=path,
                               skip_flat_paths=skip_flat,
                               recursive=recursive)

    assert_compare_errors(path=b'.', revision=b'unknown')
    assert_compare_errors(path=b'')

    # case when path is actually a file
    assert_compare_errors(path=b'sub/bar')

    # case when there is no result
    for recursive in (False, True):
        assert_compare_errors(path=b'no/such/dir', recursive=recursive)

    # sort parameter
    SortBy = GetTreeEntriesRequest.SortBy
    for recursive in (False, True):
        # status code actually depends on `recursive` in Gitaly v17.0
        assert_compare(path=b'.', recursive=recursive,
                       sort=SortBy.TREES_FIRST)

    # tree first and nested trees
    nested = sub / 'nested'
    nested.mkdir()
    (nested / 'deeper').write_text('deep thoughts')
    wrapper.commit_file('sub/nested/deeper', message='deeper')
    assert fixture.git_repo.branch_titles() == {b'branch/default': b"deeper"}

    # see comment above about update of OID mapping
    oid2git.update(oid_mapping(fixture, (
        (default_rev, b'bigfile', 'blob'),
        (default_rev, b'sub', 'tree'),
        (default_rev, b'sub/bar', 'blob'),
        (default_rev, b'sub/ba2', 'blob'),
        (default_rev, b'sub/nested', 'tree'),
        (default_rev, b'sub/nested/deeper', 'blob'),
        (default_rev, b'foo', 'blob'),
    )))
    for skip_flat in (False, True):
        assert_compare(path=b'.',
                       recursive=True,
                       skip_flat_paths=skip_flat,
                       sort=SortBy.TREES_FIRST)

    # reproduction of heptapod#191
    # According to the hints by the reporter, this is probably not the
    # smallest possible case, but encoding it exactly will spare us the need
    # to recheck it after fix.

    # let's start from scratch
    wrapper.update(b'000000000000000000000000')
    rel_subdirs = (  # subdirs with files
        'a/b/c/d',
        'a/b/c/d/e/f/g/h',
        'a/b/c/d/e/f/g/h/i/j/k/l',
        'a/b/c/d/e/f/g/h/i/j/k/l/m/n/o/p'
    )
    subdirs = [wrapper.path / subdir for subdir in rel_subdirs]

    all_subdirs = []
    deepest_missing = rel_subdirs[-1].encode()

    while True:
        all_subdirs.append(deepest_missing)
        split = deepest_missing.rsplit(b'/', 1)
        if len(split) == 1:
            break
        deepest_missing = split[0]

    test_files = [subdir / 'test' for subdir in subdirs]
    subdirs[-1].mkdir(parents=True)
    for test_file in test_files:
        test_file.write_text('hello')
    wrapper.commit(rel_paths=test_files,
                   message="Sean's reproduction",
                   branch="DrinkyBird",
                   add_remove=True)

    sean_branch = b'branch/DrinkyBird'

    oid2git.update(oid_mapping(
        fixture, ((sean_branch, subdir, 'tree')
                  for subdir in all_subdirs)
    ))
    oid2git.update(oid_mapping(
        fixture, ((sean_branch, subdir.encode() + b'/test', 'blob')
                  for subdir in rel_subdirs)
    ))
    for subdir in rel_subdirs:
        assert_compare(path=str(subdir).encode(),
                       revision=sean_branch,
                       skip_flat_paths=False,
                       )


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_get_tree_entries_pagination(gitaly_rhgitaly_comparison,
                                             hg_server):
    fixture = gitaly_rhgitaly_comparison

    wrapper = fixture.hg_repo_wrapper
    wrapper.write_commit('foo', message="Some foo")
    sub = (wrapper.path / 'sub')
    sub.mkdir()
    rel_paths = []
    # Chunk size with Gitaly is big
    nb_files = 3270
    for x in range(nb_files):
        # max file name length is 255 on most filesystems
        path = sub / ('very-' * 46 + '-long-filename-bar%04d' % x)
        path.write_text(str(x))
        rel_paths.append(path)
    # TODO OS indep for paths (actually TODO make wrapper.commit easier to
    # use, e.g., check how to make it accept patterns)
    wrapper.commit(rel_paths=rel_paths,
                   message="zebar", add_remove=True)

    def normalizer(rpc_helper, responses, **kw):
        for resp in responses:
            for entry in resp.entries:
                entry.oid = b''
            resp.pagination_cursor.next_cursor = ''

    def concat_entries(responses):
        return list(e for r in responses for e in r.entries)

    def remove_entries(response):
        del response.entries[:]

    cursor2git = {}  # hg cursor -> git cursor

    rpc_helper = fixture.rpc_helper(stub_cls=CommitServiceStub,
                                    hg_server=hg_server,
                                    method_name='GetTreeEntries',
                                    request_cls=GetTreeEntriesRequest,
                                    request_defaults=dict(
                                        revision=b'branch/default',
                                    ),
                                    streaming=True,
                                    normalizer=normalizer,
                                    chunked_fields_remover=remove_entries,
                                    chunks_concatenator=concat_entries,
                                    )

    def request_kwargs_to_git(hg_kwargs):
        """Swapping the cursor is too specific for the current RpcHelper.

        We might grow something for all paginated methods, though.
        """
        pagination = hg_kwargs.get('pagination_params')
        if pagination is None:
            return hg_kwargs

        hg_cursor = pagination.page_token
        if not hg_cursor:
            return hg_kwargs

        git_kwargs = deepcopy(hg_kwargs)
        git_kwargs['pagination_params'].page_token = cursor2git[hg_cursor]
        return git_kwargs

    rpc_helper.request_kwargs_to_git = request_kwargs_to_git

    assert_compare = rpc_helper.assert_compare
    assert_compare_errors = rpc_helper.assert_compare_errors
    assert_compare_aggregated = rpc_helper.assert_compare_aggregated

    # asking more than the expected Gitaly first chunk size (2888 entries)
    # but still less than the total
    first_resps = assert_compare_aggregated(
        path=b'sub',
        recursive=True,
        pagination_params=PaginationParameter(limit=nb_files - 2),
        compare_first_chunks=False,
    )

    git_cursor, hg_cursor = [resp.pagination_cursor.next_cursor
                             for resp in first_resps]
    assert git_cursor
    assert hg_cursor
    cursor2git[hg_cursor] = git_cursor

    # using the cursor
    assert_compare(path=b'sub',
                   recursive=True,
                   pagination_params=PaginationParameter(page_token=hg_cursor,
                                                         limit=9000))

    # negative limit means all results, and there's no cursor if no next page
    git_resp0, hg_resp0 = assert_compare_aggregated(
        path=b'sub',
        recursive=True,
        pagination_params=PaginationParameter(limit=-1),
        compare_first_chunks=False,
    )
    assert git_resp0.pagination_cursor == hg_resp0.pagination_cursor

    # case of limit=0 (empty results)
    assert_compare(path=b'sub',
                   recursive=True,
                   pagination_params=PaginationParameter(limit=0),
                   )

    # case of unknown revision and limit=0
    assert_compare_errors(path=b'.', revision=b'unknown',
                          pagination_params=PaginationParameter(limit=0),
                          skip_structured_errors_issue=156)

    # case of no pagination params
    assert_compare_aggregated(path=b'sub',
                              recursive=True,
                              compare_first_chunks=False)

    # case of a cursor that doesn't match any entry (can happen if content
    # changes between requests)
    cursor = "surely not an OID"
    cursor2git[cursor] = cursor
    rpc_helper.assert_compare_errors(path=b'sub',
                                     recursive=True,
                                     pagination_params=PaginationParameter(
                                         page_token=cursor,
                                         limit=10)
                                     )


def rev_path_messages(rev_paths):
    """Convert (revision, path) pairs into a list of `RevisionPath` messages.
    """
    return [GetBlobsRequest.RevisionPath(revision=rev, path=path)
            for rev, path in rev_paths]


@parametrize('hg_server', ('hgitaly', 'rhgitaly'))
def test_compare_get_blob_request(gitaly_rhgitaly_comparison, hg_server):
    fixture = gitaly_rhgitaly_comparison
    git_repo = fixture.git_repo

    wrapper = fixture.hg_repo_wrapper
    large_data = b'\xbe' * WRITE_BUFFER_SIZE + b'\xefdata'

    wrapper.commit_file('small', message="Small file")
    changeset = wrapper.commit_file('foo', message="Large foo",
                                    content=large_data)
    hg_sha = changeset.hex().decode()
    hg_oid = blob_oid(wrapper.repo, hg_sha, b'foo')

    default_rev = b'branch/default'

    # mirror worked
    assert git_repo.branch_titles() == {default_rev: b"Large foo"}

    oid2git = oid_mapping(fixture,
                          ((default_rev, b'foo', 'blob'),
                           (default_rev, b'small', 'blob')))
    normalizer = oid_normalizer(oid2git)

    rpc_helper = fixture.rpc_helper(stub_cls=BlobServiceStub,
                                    hg_server=hg_server,
                                    method_name='GetBlob',
                                    request_cls=GetBlobRequest,
                                    request_defaults=dict(
                                        limit=-1,
                                    ),
                                    streaming=True,
                                    normalizer=normalizer,
                                    chunked_fields_remover=remove_data_field,
                                    chunks_concatenator=concat_data_fields,
                                    )

    def request_kwargs_to_git(hg_kwargs):
        """Swapping oid is too specific for the current RpcHelper

        We might provide it with a direct `git` subprocess or with dulwich,
        though.
        """
        git_kwargs = deepcopy(hg_kwargs)
        git_kwargs['oid'] = oid2git[hg_kwargs['oid']]
        return git_kwargs

    rpc_helper.request_kwargs_to_git = request_kwargs_to_git

    rpc_helper.assert_compare(oid=hg_oid, limit=12)
    rpc_helper.assert_compare(oid=hg_oid, limit=0)
    for cs, path in (('12de34ad' * 5, b'foo'),
                     (ZERO_SHA_STR, b'irrelevant/path'),
                     (hg_sha, b'unknown-file'),
                     ):
        unknown_oid = blob_oid(wrapper.repo, cs, path)
        oid2git[unknown_oid] = unknown_oid
        rpc_helper.assert_compare(oid=unknown_oid)

    rpc_helper.assert_compare_aggregated(oid=hg_oid,
                                         compare_first_chunks=False)

    oid2git[''] = ''
    rpc_helper.assert_compare_errors(repository=None, oid='',
                                     # details differ just by capitalization
                                     # as of this writing
                                     same_details=False,
                                     )
    rpc_helper.assert_compare_errors(oid='')
    bogus_repos = (Repository(storage_name='unknown',
                              relative_path='some_path'),
                   Repository(storage_name=fixture.gitaly_repo.storage_name,
                              relative_path='does/not/exist'),
                   None,
                   )

    for repo in bogus_repos:
        rpc_helper.assert_compare_errors(repository=repo,
                                         oid=hg_oid,
                                         same_details=False)

    # now with GetBlobs
    rpc_helper = fixture.rpc_helper(stub_cls=BlobServiceStub,
                                    hg_server=hg_server,
                                    method_name='GetBlobs',
                                    request_cls=GetBlobsRequest,
                                    request_defaults=dict(
                                        limit=-1,
                                    ),
                                    streaming=True,
                                    normalizer=normalizer,
                                    chunked_fields_remover=remove_data_field,
                                    chunks_concatenator=concat_data_fields,
                                    )

    rev_paths = rev_path_messages(((b'branch/default', b'small'),
                                   (b'branch/default', b'does-not-exist'),
                                   (b'no-such-revision', b'small'),
                                   (ZERO_SHA_STR, b'irrelevant/path'),
                                   ))
    rpc_helper.assert_compare(revision_paths=rev_paths)
    # with limits (the limit is per file)
    # the case of limit=0 is important, being used by the Rails app to
    # obtain metadata only (see e.g., `BlobService.get_blob_types`).
    for limit in (0, 3):
        rpc_helper.assert_compare(revision_paths=rev_paths, limit=limit)

    # chunking in get_blobs, again non-deterministic for Gitaly
    rev_paths = rev_path_messages(((b'branch/default', b'small'),
                                   (b'branch/default', b'foo'),
                                   ))
    rpc_helper.assert_compare_aggregated(revision_paths=rev_paths)

    # corner cases
    for repo in bogus_repos:
        rpc_helper.assert_compare_errors(repository=repo,
                                         revision_paths=rev_paths,
                                         same_details=False)


def test_check_get_blobs_generated(gitaly_comparison):
    fixture = gitaly_comparison
    git_repo = fixture.git_repo

    wrapper = fixture.hg_repo_wrapper

    sub = (wrapper.path / 'sub')
    sub.mkdir()

    cargo_path = b'sub/Cargo.lock'
    (wrapper.path / cargo_path.decode()).write_text("Content will not matter")
    (wrapper.path / 'foo').write_text("foo")

    stub_path = b'commit_pb2.py'
    import hgitaly.stub.commit_pb2 as stub_exmp
    shutil.copy(stub_exmp.__file__, wrapper.path / stub_path.decode())

    changeset = wrapper.commit(rel_paths=[cargo_path, stub_path, 'foo'],
                               add_remove=True,
                               message="Bunch of files")
    hg_sha = changeset.hex().decode()

    cargo_oid = blob_oid(wrapper.repo, hg_sha, cargo_path)
    foo_oid = blob_oid(wrapper.repo, hg_sha, b'foo')
    stub_oid = blob_oid(wrapper.repo, hg_sha, stub_path)

    default_rev = b'branch/default'

    # mirror worked
    assert git_repo.branch_titles() == {default_rev: b"Bunch of files"}

    oid2git = oid_mapping(fixture,
                          ((default_rev, cargo_path, 'blob'),
                           (default_rev, stub_path, 'blob'),
                           (default_rev, b'foo', 'blob')))

    def blobs_normalizer(rpc_helper, responses, vcs='hg'):
        if vcs != 'hg':
            return
        for resp in responses:
            for blob in resp.blobs:
                blob.revision = oid2git[blob.revision.decode()].encode()

    rpc_helper = fixture.rpc_helper(stub_cls=AnalysisServiceStub,
                                    method_name='CheckBlobsGenerated',
                                    request_cls=CheckBlobsGeneratedRequest,
                                    streaming=True,
                                    streaming_request_field='blobs',
                                    normalizer=blobs_normalizer,
                                    )

    Blob = CheckBlobsGeneratedRequest.Blob

    def request_kwargs_to_git(hg_kwargs):
        """Swapping oid is too specific for the current RpcHelper

        We might provide it with a direct `git` subprocess or with dulwich,
        though.
        """
        git_kwargs = deepcopy(hg_kwargs)
        for hg_blob, git_blob in zip(hg_kwargs['blobs'], git_kwargs['blobs']):
            git_blob.revision = oid2git[hg_blob.revision.decode()].encode()
        return git_kwargs

    rpc_helper.request_kwargs_to_git = request_kwargs_to_git

    rpc_helper.assert_compare(
        blobs=[Blob(revision=cargo_oid.encode(), path=cargo_path),
               Blob(revision=foo_oid.encode(), path=b'foo'),
               Blob(revision=stub_oid.encode(), path=stub_path),
               ]
    )

    rpc_helper.assert_compare(blobs=[Blob(revision=cargo_oid.encode(),
                                          path=cargo_path)])

    # Testing the commit_id:path syntax actually used by the Rails app
    # (the bytes/str dance is annoying but is not the point)
    git_sha = rpc_helper.hg2git(hg_sha).decode()
    cargo_rev = ':'.join((hg_sha, cargo_path.decode()))
    cargo_git_rev = ':'.join((git_sha, cargo_path.decode()))
    oid2git[cargo_rev] = cargo_git_rev
    rpc_helper.assert_compare(
        blobs=[Blob(revision=cargo_rev.encode(), path=cargo_path)]
    )

    # formally valid, but missing oids
    missing_path = b'unknown-file'
    arbitrary_sha1 = '12fe34ca' * 5
    missing_oid = blob_oid(wrapper.repo, hg_sha, missing_path)

    oid2git[missing_oid] = arbitrary_sha1
    rpc_helper.assert_compare_errors(
        blobs=[Blob(revision=missing_oid.encode(), path=missing_path)])

    missing_oid = blob_oid(wrapper.repo, arbitrary_sha1, missing_path)
    oid2git[missing_oid] = arbitrary_sha1
    rpc_helper.assert_compare_errors(
        blobs=[Blob(revision=missing_oid.encode(), path=missing_path)])

    # invalid oid
    missing_oid = '123'
    oid2git[missing_oid] = missing_oid
    rpc_helper.assert_compare_errors(
        blobs=[Blob(revision=missing_oid.encode(), path=missing_path)])



# Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import json
import subprocess

from grpc import StatusCode
from importlib_resources import files


def read_spdx_licenses(path):
    with path.open() as licf:
        return {lic['licenseId'].lower(): lic
                for lic in json.load(licf)['licenses']}


DATA_DIR = files(__name__)
SPDX_LICENSES = read_spdx_licenses(DATA_DIR.joinpath('spdx-licenses.json'))
"""Official SPDX licenses file, by lowercase id.

The `spdx` Python package is clearly outdated,
the `spdx-tools` Python package is so complex it is not even clear in
a few minutes if it contains the information we need.
"""

LICENSE_ID_TO_NICK_NAMES = {
    "agpl-3.0": "GNU AGPLv3",
    "lgpl-3.0": "GNU LGPLv3",
    "bsd-3-clause-clear": "Clear BSD",
    "odbl-1.0": "ODbL",
    "ncsa": "UIUC/NCSA",
    "lgpl-2.1": "GNU LGPLv2.1",
    "gpl-3.0": "GNU GPLv3",
    "gpl-2.0": "GNU GPLv2",
}
"""Same as in Gitaly's implementation, we need a mapping of nicknames

See commment (as of Gitaly 16.2) in
/internal/gitaly/service/repository/license.go

It is not fully clear what this nickname is supposed to be, as it is not
a SPDX concept, could be GitLab-specific or carried over from a previous
system (Ruby gem from GitHub?)
"""

DEPRECATED_PREFIX = 'deprecated_'

COMMON_LICENSE_FILES = ('LICENSE', 'COPYING')


def license_content(name):
    """Return the full test content of one of the licenses we ship.

    Intendend for gRPC tests and Gitaly comparison tests.
    """
    return (DATA_DIR / (name + '.sample')).read_text()


def strip_deprecated_prefix(name):
    if name.startswith(DEPRECATED_PREFIX):
        return name[len(DEPRECATED_PREFIX):]
    return name


def match_sort_key(match):
    return (match['confidence'],
            strip_deprecated_prefix(match['license']).lower())


def detect_license(bin_path, path, context):
    """Return the full language stats working directory at given path.

    :param str bin_path: path to the Go `license-detector` executable
    :param path: the directory to analyze as a :class:`pathlib.Path` instance.
    """
    try:
        detector = subprocess.run(
            (bin_path, '-f', 'json', path),
            encoding='utf-8',
            check=True,
            capture_output=True,
        )
    except subprocess.CalledProcessError as exc:
        # in Gitaly 15.5, the resulting error code is 'INTERNAL', with
        # details 'language stats: waiting for linguist: exit status %d'
        context.abort(
            StatusCode.INTERNAL,
            "detect_license: Go `license-detector` error "
            "status code=%d stderr=%r" % (
                exc.returncode, exc.stderr))
    except FileNotFoundError:
        context.abort(
            StatusCode.INTERNAL,
            "detect_license: Go `license-detector` executable '%s' "
            "not found" % bin_path)
    except PermissionError:
        # Gitaly 15.5 behaves similarly in this case as in the not found case,
        # only with exit status 126 in details (same as in a shell, again)
        context.abort(
            StatusCode.INTERNAL,
            "detect_license: permission denied to run "
            "Go `license-detector` executable '%s'" % bin_path)

    detected = json.loads(detector.stdout)[0]
    error = detected.get('error')
    if error is not None:
        if error != 'no license file was found':
            context.abort(StatusCode.INTERNAL, "detect_license: " + error)

        for fname in COMMON_LICENSE_FILES:
            if (path / fname).exists():
                return dict(spdx_id='other',
                            full_name='Other',
                            file=fname,
                            nickname='LICENSE')
        return None

    # TODO same as Gitaly, sort by identifier, same reason: reproducibility
    # in case of equal confidence score.
    matches = detected['matches']
    matches.sort(key=match_sort_key)
    detected = matches[0]
    spdx_id = strip_deprecated_prefix(detected['license']).lower()
    detected['spdx_id'] = spdx_id

    spdx = SPDX_LICENSES.get(spdx_id)
    if spdx is not None:  # TODO what if not found?
        detected['full_name'] = spdx['name']
        # this is far from correct (e.g., it returns the text of GPL-2.0
        # for GPL-2.0+) but this is what Gitaly seems to do.
        detected['url'] = spdx['seeAlso'][0]

    detected['nickname'] = LICENSE_ID_TO_NICK_NAMES.get(spdx_id, "")
    return detected



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import repository_pb2 as repository__pb2


class RepositoryServiceStub(object):
    """RepositoryService is a service providing RPCs accessing repositories as a whole.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RepositoryExists = channel.unary_unary(
                '/gitaly.RepositoryService/RepositoryExists',
                request_serializer=repository__pb2.RepositoryExistsRequest.SerializeToString,
                response_deserializer=repository__pb2.RepositoryExistsResponse.FromString,
                )
        self.RepositorySize = channel.unary_unary(
                '/gitaly.RepositoryService/RepositorySize',
                request_serializer=repository__pb2.RepositorySizeRequest.SerializeToString,
                response_deserializer=repository__pb2.RepositorySizeResponse.FromString,
                )
        self.RepositoryInfo = channel.unary_unary(
                '/gitaly.RepositoryService/RepositoryInfo',
                request_serializer=repository__pb2.RepositoryInfoRequest.SerializeToString,
                response_deserializer=repository__pb2.RepositoryInfoResponse.FromString,
                )
        self.ObjectsSize = channel.stream_unary(
                '/gitaly.RepositoryService/ObjectsSize',
                request_serializer=repository__pb2.ObjectsSizeRequest.SerializeToString,
                response_deserializer=repository__pb2.ObjectsSizeResponse.FromString,
                )
        self.ObjectFormat = channel.unary_unary(
                '/gitaly.RepositoryService/ObjectFormat',
                request_serializer=repository__pb2.ObjectFormatRequest.SerializeToString,
                response_deserializer=repository__pb2.ObjectFormatResponse.FromString,
                )
        self.ApplyGitattributes = channel.unary_unary(
                '/gitaly.RepositoryService/ApplyGitattributes',
                request_serializer=repository__pb2.ApplyGitattributesRequest.SerializeToString,
                response_deserializer=repository__pb2.ApplyGitattributesResponse.FromString,
                )
        self.FetchRemote = channel.unary_unary(
                '/gitaly.RepositoryService/FetchRemote',
                request_serializer=repository__pb2.FetchRemoteRequest.SerializeToString,
                response_deserializer=repository__pb2.FetchRemoteResponse.FromString,
                )
        self.CreateRepository = channel.unary_unary(
                '/gitaly.RepositoryService/CreateRepository',
                request_serializer=repository__pb2.CreateRepositoryRequest.SerializeToString,
                response_deserializer=repository__pb2.CreateRepositoryResponse.FromString,
                )
        self.GetArchive = channel.unary_stream(
                '/gitaly.RepositoryService/GetArchive',
                request_serializer=repository__pb2.GetArchiveRequest.SerializeToString,
                response_deserializer=repository__pb2.GetArchiveResponse.FromString,
                )
        self.HasLocalBranches = channel.unary_unary(
                '/gitaly.RepositoryService/HasLocalBranches',
                request_serializer=repository__pb2.HasLocalBranchesRequest.SerializeToString,
                response_deserializer=repository__pb2.HasLocalBranchesResponse.FromString,
                )
        self.FetchSourceBranch = channel.unary_unary(
                '/gitaly.RepositoryService/FetchSourceBranch',
                request_serializer=repository__pb2.FetchSourceBranchRequest.SerializeToString,
                response_deserializer=repository__pb2.FetchSourceBranchResponse.FromString,
                )
        self.Fsck = channel.unary_unary(
                '/gitaly.RepositoryService/Fsck',
                request_serializer=repository__pb2.FsckRequest.SerializeToString,
                response_deserializer=repository__pb2.FsckResponse.FromString,
                )
        self.WriteRef = channel.unary_unary(
                '/gitaly.RepositoryService/WriteRef',
                request_serializer=repository__pb2.WriteRefRequest.SerializeToString,
                response_deserializer=repository__pb2.WriteRefResponse.FromString,
                )
        self.FindMergeBase = channel.unary_unary(
                '/gitaly.RepositoryService/FindMergeBase',
                request_serializer=repository__pb2.FindMergeBaseRequest.SerializeToString,
                response_deserializer=repository__pb2.FindMergeBaseResponse.FromString,
                )
        self.CreateFork = channel.unary_unary(
                '/gitaly.RepositoryService/CreateFork',
                request_serializer=repository__pb2.CreateForkRequest.SerializeToString,
                response_deserializer=repository__pb2.CreateForkResponse.FromString,
                )
        self.CreateRepositoryFromURL = channel.unary_unary(
                '/gitaly.RepositoryService/CreateRepositoryFromURL',
                request_serializer=repository__pb2.CreateRepositoryFromURLRequest.SerializeToString,
                response_deserializer=repository__pb2.CreateRepositoryFromURLResponse.FromString,
                )
        self.CreateBundle = channel.unary_stream(
                '/gitaly.RepositoryService/CreateBundle',
                request_serializer=repository__pb2.CreateBundleRequest.SerializeToString,
                response_deserializer=repository__pb2.CreateBundleResponse.FromString,
                )
        self.CreateBundleFromRefList = channel.stream_stream(
                '/gitaly.RepositoryService/CreateBundleFromRefList',
                request_serializer=repository__pb2.CreateBundleFromRefListRequest.SerializeToString,
                response_deserializer=repository__pb2.CreateBundleFromRefListResponse.FromString,
                )
        self.GenerateBundleURI = channel.unary_unary(
                '/gitaly.RepositoryService/GenerateBundleURI',
                request_serializer=repository__pb2.GenerateBundleURIRequest.SerializeToString,
                response_deserializer=repository__pb2.GenerateBundleURIResponse.FromString,
                )
        self.FetchBundle = channel.stream_unary(
                '/gitaly.RepositoryService/FetchBundle',
                request_serializer=repository__pb2.FetchBundleRequest.SerializeToString,
                response_deserializer=repository__pb2.FetchBundleResponse.FromString,
                )
        self.CreateRepositoryFromBundle = channel.stream_unary(
                '/gitaly.RepositoryService/CreateRepositoryFromBundle',
                request_serializer=repository__pb2.CreateRepositoryFromBundleRequest.SerializeToString,
                response_deserializer=repository__pb2.CreateRepositoryFromBundleResponse.FromString,
                )
        self.GetConfig = channel.unary_stream(
                '/gitaly.RepositoryService/GetConfig',
                request_serializer=repository__pb2.GetConfigRequest.SerializeToString,
                response_deserializer=repository__pb2.GetConfigResponse.FromString,
                )
        self.FindLicense = channel.unary_unary(
                '/gitaly.RepositoryService/FindLicense',
                request_serializer=repository__pb2.FindLicenseRequest.SerializeToString,
                response_deserializer=repository__pb2.FindLicenseResponse.FromString,
                )
        self.GetInfoAttributes = channel.unary_stream(
                '/gitaly.RepositoryService/GetInfoAttributes',
                request_serializer=repository__pb2.GetInfoAttributesRequest.SerializeToString,
                response_deserializer=repository__pb2.GetInfoAttributesResponse.FromString,
                )
        self.CalculateChecksum = channel.unary_unary(
                '/gitaly.RepositoryService/CalculateChecksum',
                request_serializer=repository__pb2.CalculateChecksumRequest.SerializeToString,
                response_deserializer=repository__pb2.CalculateChecksumResponse.FromString,
                )
        self.GetSnapshot = channel.unary_stream(
                '/gitaly.RepositoryService/GetSnapshot',
                request_serializer=repository__pb2.GetSnapshotRequest.SerializeToString,
                response_deserializer=repository__pb2.GetSnapshotResponse.FromString,
                )
        self.CreateRepositoryFromSnapshot = channel.unary_unary(
                '/gitaly.RepositoryService/CreateRepositoryFromSnapshot',
                request_serializer=repository__pb2.CreateRepositoryFromSnapshotRequest.SerializeToString,
                response_deserializer=repository__pb2.CreateRepositoryFromSnapshotResponse.FromString,
                )
        self.GetRawChanges = channel.unary_stream(
                '/gitaly.RepositoryService/GetRawChanges',
                request_serializer=repository__pb2.GetRawChangesRequest.SerializeToString,
                response_deserializer=repository__pb2.GetRawChangesResponse.FromString,
                )
        self.SearchFilesByContent = channel.unary_stream(
                '/gitaly.RepositoryService/SearchFilesByContent',
                request_serializer=repository__pb2.SearchFilesByContentRequest.SerializeToString,
                response_deserializer=repository__pb2.SearchFilesByContentResponse.FromString,
                )
        self.SearchFilesByName = channel.unary_stream(
                '/gitaly.RepositoryService/SearchFilesByName',
                request_serializer=repository__pb2.SearchFilesByNameRequest.SerializeToString,
                response_deserializer=repository__pb2.SearchFilesByNameResponse.FromString,
                )
        self.RestoreCustomHooks = channel.stream_unary(
                '/gitaly.RepositoryService/RestoreCustomHooks',
                request_serializer=repository__pb2.RestoreCustomHooksRequest.SerializeToString,
                response_deserializer=repository__pb2.RestoreCustomHooksResponse.FromString,
                )
        self.SetCustomHooks = channel.stream_unary(
                '/gitaly.RepositoryService/SetCustomHooks',
                request_serializer=repository__pb2.SetCustomHooksRequest.SerializeToString,
                response_deserializer=repository__pb2.SetCustomHooksResponse.FromString,
                )
        self.BackupCustomHooks = channel.unary_stream(
                '/gitaly.RepositoryService/BackupCustomHooks',
                request_serializer=repository__pb2.BackupCustomHooksRequest.SerializeToString,
                response_deserializer=repository__pb2.BackupCustomHooksResponse.FromString,
                )
        self.GetCustomHooks = channel.unary_stream(
                '/gitaly.RepositoryService/GetCustomHooks',
                request_serializer=repository__pb2.GetCustomHooksRequest.SerializeToString,
                response_deserializer=repository__pb2.GetCustomHooksResponse.FromString,
                )
        self.GetObjectDirectorySize = channel.unary_unary(
                '/gitaly.RepositoryService/GetObjectDirectorySize',
                request_serializer=repository__pb2.GetObjectDirectorySizeRequest.SerializeToString,
                response_deserializer=repository__pb2.GetObjectDirectorySizeResponse.FromString,
                )
        self.RemoveRepository = channel.unary_unary(
                '/gitaly.RepositoryService/RemoveRepository',
                request_serializer=repository__pb2.RemoveRepositoryRequest.SerializeToString,
                response_deserializer=repository__pb2.RemoveRepositoryResponse.FromString,
                )
        self.ReplicateRepository = channel.unary_unary(
                '/gitaly.RepositoryService/ReplicateRepository',
                request_serializer=repository__pb2.ReplicateRepositoryRequest.SerializeToString,
                response_deserializer=repository__pb2.ReplicateRepositoryResponse.FromString,
                )
        self.OptimizeRepository = channel.unary_unary(
                '/gitaly.RepositoryService/OptimizeRepository',
                request_serializer=repository__pb2.OptimizeRepositoryRequest.SerializeToString,
                response_deserializer=repository__pb2.OptimizeRepositoryResponse.FromString,
                )
        self.PruneUnreachableObjects = channel.unary_unary(
                '/gitaly.RepositoryService/PruneUnreachableObjects',
                request_serializer=repository__pb2.PruneUnreachableObjectsRequest.SerializeToString,
                response_deserializer=repository__pb2.PruneUnreachableObjectsResponse.FromString,
                )
        self.BackupRepository = channel.unary_unary(
                '/gitaly.RepositoryService/BackupRepository',
                request_serializer=repository__pb2.BackupRepositoryRequest.SerializeToString,
                response_deserializer=repository__pb2.BackupRepositoryResponse.FromString,
                )
        self.RestoreRepository = channel.unary_unary(
                '/gitaly.RepositoryService/RestoreRepository',
                request_serializer=repository__pb2.RestoreRepositoryRequest.SerializeToString,
                response_deserializer=repository__pb2.RestoreRepositoryResponse.FromString,
                )
        self.GetFileAttributes = channel.unary_unary(
                '/gitaly.RepositoryService/GetFileAttributes',
                request_serializer=repository__pb2.GetFileAttributesRequest.SerializeToString,
                response_deserializer=repository__pb2.GetFileAttributesResponse.FromString,
                )
        self.FastExport = channel.unary_stream(
                '/gitaly.RepositoryService/FastExport',
                request_serializer=repository__pb2.FastExportRequest.SerializeToString,
                response_deserializer=repository__pb2.FastExportResponse.FromString,
                )


class RepositoryServiceServicer(object):
    """RepositoryService is a service providing RPCs accessing repositories as a whole.
    """

    def RepositoryExists(self, request, context):
        """RepositoryExists returns whether a given repository exists.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RepositorySize(self, request, context):
        """RepositorySize returns information on the complete on-disk repository size. If you need more
        detailed information about the size of various sub-structures you should instead use the
        repositoryInfo RPC.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RepositoryInfo(self, request, context):
        """RepositoryInfo returns detailed information about a repository and its data structures.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ObjectsSize(self, request_iterator, context):
        """ObjectsSize calculates the total on-disk object size of reachable objects in bytes. In contrast
        to RepositorySize and RepositoryInfo, this RPC performs a graph walk of the specified revisions
        and will thus return an accurate view of how large the accumulated on-disk size of reachable
        objects is.

        As this RPC needs to perform a revision walk, it is significantly more expensive than the RPCs
        which simply check the size of on-disk data structures. On the other hand, it allows the caller
        to accurately compute the size of objects in a way that is at least somewhat detached from the
        on-disk representation:

        - Objects which exist in multiple packfiles will not be double-counted.
        - Objects which aren't reachable will not be accounted for.
        - It is possible to only account for a subset of references, e.g. only those that an admin
        would have direct control over.

        It is thus recommended to use this RPC whenever you want to calculate sizes which will end up
        being shown to the user.

        Note that the size is still bound to change when repositories are getting repacked and thus
        cannot be considered to be stable. This is because the on-disk size of any object can change
        depending on how Git decides to deltify it in a packfile. Thus, when a repack would cause a
        different delta base to be picked, the actual on-disk size of any given object may change.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ObjectFormat(self, request, context):
        """ObjectFormat determines the object format that is being used by the repository.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ApplyGitattributes(self, request, context):
        """ApplyGitattributes writes the attributes from the given revision to info/attributes.
        This RPC will be removed in 17.0.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FetchRemote(self, request, context):
        """FetchRemote fetches references from a remote repository into the local
        repository. The remote can be fetched via HTTP or SSH depending on the
        request options provided.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateRepository(self, request, context):
        """CreateRepository creates a new empty repository.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetArchive(self, request, context):
        """GetArchive produces and returns an archive of a repository.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HasLocalBranches(self, request, context):
        """HasLocalBranches returns whether the given repo contains any branches.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FetchSourceBranch(self, request, context):
        """FetchSourceBranch fetches a branch from a second (potentially remote)
        repository into the given repository.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Fsck(self, request, context):
        """Fsck checks the repository for consistency via git-fsck(1). This can be used to check for
        repository corruption.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def WriteRef(self, request, context):
        """WriteRef creates or updates a ref in a repository to point to a new value.
        If the request contains object revisions that do not exist in the repository, the RPC returns
        the NOT_FOUND code with the ReferenceNotFoundError detail.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FindMergeBase(self, request, context):
        """FindMergeBase returns the best common ancestor between two or more commits. Consult the man
        pages of git-merge-base(1) for more information on how this is calculated.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateFork(self, request, context):
        """CreateFork creates a new repository from a specific source repository. This new repository will
        have the same branches and tags as the source repository. Internal references will not be
        recreated in the forked repository.

        all objects of the source repository will be duplicated, that is there are no space savings by
        creating the repository like this. The newly created repository does not join the object pool
        of the source repository, if there is any.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateRepositoryFromURL(self, request, context):
        """CreateRepositoryFromURL creates a new repo and seeds it with the contents of an existing Git repo
        reachable at the provided URL.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateBundle(self, request, context):
        """CreateBundle creates a bundle from all refs
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateBundleFromRefList(self, request_iterator, context):
        """CreateBundleFromRefList creates a bundle from a stream of ref patterns.
        When the bundle would be empty the FailedPrecondition error code is returned.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GenerateBundleURI(self, request, context):
        """GenerateBundleURI generates a bundle on the server for bundle-URI use.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FetchBundle(self, request_iterator, context):
        """FetchBundle fetches references from a bundle into the local repository.
        refs will be mirrored to the target repository with the refspec
        "+refs/*:refs/*" and refs that do not exist in the bundle will be removed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateRepositoryFromBundle(self, request_iterator, context):
        """CreateRepositoryFromBundle creates a Git repository at the specified storage and path, if it
        does not already exist, from the provided Git bundle.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetConfig(self, request, context):
        """GetConfig reads the target repository's gitconfig and streams its contents
        back. Returns a NotFound error in case no gitconfig was found.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FindLicense(self, request, context):
        """FindLicense looks in the given repository and attempts to detect all the
        details about the license used in the repository.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetInfoAttributes(self, request, context):
        """GetInfoAttributes reads the contents from info/attributes.
        This RPC will be removed in 17.0.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CalculateChecksum(self, request, context):
        """CalculateChecksum returns a checksum of the repository by hashing its references. Refs
        outside of well-known namespaces are not considered when computing the checksum.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSnapshot(self, request, context):
        """GetSnapshot returns a snapshot of the repository. A snapshot comprises all Git references
        and objects required to recreate the state of a repository at a point in time.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateRepositoryFromSnapshot(self, request, context):
        """CreateRepositoryFromSnapshot creates a new repository based on a snapshot created with
        the GetSnapshot RPC. The snapshot is fetched via HTTP.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetRawChanges(self, request, context):
        """GetRawChanges returns metadata in raw format on the changes between two revisions.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SearchFilesByContent(self, request, context):
        """SearchFilesByContent searches files in the repository using the provided grep pattern.
        For each result, the matched line is returned along with the two previous and next lines.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SearchFilesByName(self, request, context):
        """SearchFilesByName searches files in the repository based on its name and an
        optional filter.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RestoreCustomHooks(self, request_iterator, context):
        """RestoreCustomHooks sets the git hooks for a repository. The hooks are sent
        in a tar archive containing a `custom_hooks` directory. This directory is
        ultimately extracted to the repository.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetCustomHooks(self, request_iterator, context):
        """SetCustomHooks sets the git hooks for a repository. The hooks are sent in a
        tar archive containing a `custom_hooks` directory (i.e. the response from the
        GetCustomHooksResponse RPC. This directory will be extracted into the repository.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BackupCustomHooks(self, request, context):
        """BackupCustomHooks fetches the git hooks for a repository. The hooks are
        sent in a tar archive containing a `custom_hooks` directory. If no hooks
        are present in the repository, the response will have no data.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetCustomHooks(self, request, context):
        """GetCustomHooks fetches the git hooks for a repository. The hooks are sent
        in a tar archive containing a `custom_hooks` directory. If no hooks are
        present in the repository, the response will have no data.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetObjectDirectorySize(self, request, context):
        """GetObjectDirectorySize returns the size in kibibytes of the object directory of a repository.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemoveRepository(self, request, context):
        """RemoveRepository will move the repository to `+gitaly/tmp/<relative_path>_removed` and
        eventually remove it. This ensures that even on networked filesystems the
        data is actually removed even if there's someone still handling the data.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReplicateRepository(self, request, context):
        """ReplicateRepository replicates data from a source repository to target repository. On the
        target repository, this operation ensures synchronization of the following components:

        - Git config
        - Git attributes
        - Custom Git hooks,
        - References and objects
        - (Optional) Object deduplication network membership
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OptimizeRepository(self, request, context):
        """OptimizeRepository performs all maintenance tasks in a repository to keep
        it in an efficient state. It cleans up stale data, repacks objects,
        updates auxiliary caches like commit-graphs and packs references. The
        optimizations performed are based on heuristics and will adapt to the
        repository's size. This RPC call is designed as a black-box such that
        Gitaly has complete control of the on-disk state of repositories.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PruneUnreachableObjects(self, request, context):
        """PruneUnreachableObjects will prune all objects which aren't reachable from
        the repository's current set of references. Because pruning can only
        happen for objects which aren't packed, you are required to first run
        OptimizeRepository to explode any unreachable objects into loose objects.

        Furthermore, this RPC call has a grace period of 30 minutes: any
        unreachable loose objects must not have been accessed or modified in the
        last 30 minutes. This is a hard requirement to avoid repository corruption.

        to make proper use of this RPC you thus need to call OptimizeRepository,
        wait 30 minutes, and then call PruneUnreachableObjects.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BackupRepository(self, request, context):
        """BackupRepository creates a full or incremental backup streamed directly to
        object-storage. The backup is created synchronously. The destination must
        be configured in config.backup.go_cloud_url
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RestoreRepository(self, request, context):
        """RestoreRepository restores a backup streamed directly from object-storage.
        The repository is restored synchronously. The source object-storage must
        be configured in config.backup.go_cloud_url
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFileAttributes(self, request, context):
        """GetFileAttributes queries given file attributes as specified in .gitattributes file
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FastExport(self, request, context):
        """FastExport runs git-fast-export on the repository, streaming the data back through the response
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RepositoryServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RepositoryExists': grpc.unary_unary_rpc_method_handler(
                    servicer.RepositoryExists,
                    request_deserializer=repository__pb2.RepositoryExistsRequest.FromString,
                    response_serializer=repository__pb2.RepositoryExistsResponse.SerializeToString,
            ),
            'RepositorySize': grpc.unary_unary_rpc_method_handler(
                    servicer.RepositorySize,
                    request_deserializer=repository__pb2.RepositorySizeRequest.FromString,
                    response_serializer=repository__pb2.RepositorySizeResponse.SerializeToString,
            ),
            'RepositoryInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.RepositoryInfo,
                    request_deserializer=repository__pb2.RepositoryInfoRequest.FromString,
                    response_serializer=repository__pb2.RepositoryInfoResponse.SerializeToString,
            ),
            'ObjectsSize': grpc.stream_unary_rpc_method_handler(
                    servicer.ObjectsSize,
                    request_deserializer=repository__pb2.ObjectsSizeRequest.FromString,
                    response_serializer=repository__pb2.ObjectsSizeResponse.SerializeToString,
            ),
            'ObjectFormat': grpc.unary_unary_rpc_method_handler(
                    servicer.ObjectFormat,
                    request_deserializer=repository__pb2.ObjectFormatRequest.FromString,
                    response_serializer=repository__pb2.ObjectFormatResponse.SerializeToString,
            ),
            'ApplyGitattributes': grpc.unary_unary_rpc_method_handler(
                    servicer.ApplyGitattributes,
                    request_deserializer=repository__pb2.ApplyGitattributesRequest.FromString,
                    response_serializer=repository__pb2.ApplyGitattributesResponse.SerializeToString,
            ),
            'FetchRemote': grpc.unary_unary_rpc_method_handler(
                    servicer.FetchRemote,
                    request_deserializer=repository__pb2.FetchRemoteRequest.FromString,
                    response_serializer=repository__pb2.FetchRemoteResponse.SerializeToString,
            ),
            'CreateRepository': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateRepository,
                    request_deserializer=repository__pb2.CreateRepositoryRequest.FromString,
                    response_serializer=repository__pb2.CreateRepositoryResponse.SerializeToString,
            ),
            'GetArchive': grpc.unary_stream_rpc_method_handler(
                    servicer.GetArchive,
                    request_deserializer=repository__pb2.GetArchiveRequest.FromString,
                    response_serializer=repository__pb2.GetArchiveResponse.SerializeToString,
            ),
            'HasLocalBranches': grpc.unary_unary_rpc_method_handler(
                    servicer.HasLocalBranches,
                    request_deserializer=repository__pb2.HasLocalBranchesRequest.FromString,
                    response_serializer=repository__pb2.HasLocalBranchesResponse.SerializeToString,
            ),
            'FetchSourceBranch': grpc.unary_unary_rpc_method_handler(
                    servicer.FetchSourceBranch,
                    request_deserializer=repository__pb2.FetchSourceBranchRequest.FromString,
                    response_serializer=repository__pb2.FetchSourceBranchResponse.SerializeToString,
            ),
            'Fsck': grpc.unary_unary_rpc_method_handler(
                    servicer.Fsck,
                    request_deserializer=repository__pb2.FsckRequest.FromString,
                    response_serializer=repository__pb2.FsckResponse.SerializeToString,
            ),
            'WriteRef': grpc.unary_unary_rpc_method_handler(
                    servicer.WriteRef,
                    request_deserializer=repository__pb2.WriteRefRequest.FromString,
                    response_serializer=repository__pb2.WriteRefResponse.SerializeToString,
            ),
            'FindMergeBase': grpc.unary_unary_rpc_method_handler(
                    servicer.FindMergeBase,
                    request_deserializer=repository__pb2.FindMergeBaseRequest.FromString,
                    response_serializer=repository__pb2.FindMergeBaseResponse.SerializeToString,
            ),
            'CreateFork': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateFork,
                    request_deserializer=repository__pb2.CreateForkRequest.FromString,
                    response_serializer=repository__pb2.CreateForkResponse.SerializeToString,
            ),
            'CreateRepositoryFromURL': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateRepositoryFromURL,
                    request_deserializer=repository__pb2.CreateRepositoryFromURLRequest.FromString,
                    response_serializer=repository__pb2.CreateRepositoryFromURLResponse.SerializeToString,
            ),
            'CreateBundle': grpc.unary_stream_rpc_method_handler(
                    servicer.CreateBundle,
                    request_deserializer=repository__pb2.CreateBundleRequest.FromString,
                    response_serializer=repository__pb2.CreateBundleResponse.SerializeToString,
            ),
            'CreateBundleFromRefList': grpc.stream_stream_rpc_method_handler(
                    servicer.CreateBundleFromRefList,
                    request_deserializer=repository__pb2.CreateBundleFromRefListRequest.FromString,
                    response_serializer=repository__pb2.CreateBundleFromRefListResponse.SerializeToString,
            ),
            'GenerateBundleURI': grpc.unary_unary_rpc_method_handler(
                    servicer.GenerateBundleURI,
                    request_deserializer=repository__pb2.GenerateBundleURIRequest.FromString,
                    response_serializer=repository__pb2.GenerateBundleURIResponse.SerializeToString,
            ),
            'FetchBundle': grpc.stream_unary_rpc_method_handler(
                    servicer.FetchBundle,
                    request_deserializer=repository__pb2.FetchBundleRequest.FromString,
                    response_serializer=repository__pb2.FetchBundleResponse.SerializeToString,
            ),
            'CreateRepositoryFromBundle': grpc.stream_unary_rpc_method_handler(
                    servicer.CreateRepositoryFromBundle,
                    request_deserializer=repository__pb2.CreateRepositoryFromBundleRequest.FromString,
                    response_serializer=repository__pb2.CreateRepositoryFromBundleResponse.SerializeToString,
            ),
            'GetConfig': grpc.unary_stream_rpc_method_handler(
                    servicer.GetConfig,
                    request_deserializer=repository__pb2.GetConfigRequest.FromString,
                    response_serializer=repository__pb2.GetConfigResponse.SerializeToString,
            ),
            'FindLicense': grpc.unary_unary_rpc_method_handler(
                    servicer.FindLicense,
                    request_deserializer=repository__pb2.FindLicenseRequest.FromString,
                    response_serializer=repository__pb2.FindLicenseResponse.SerializeToString,
            ),
            'GetInfoAttributes': grpc.unary_stream_rpc_method_handler(
                    servicer.GetInfoAttributes,
                    request_deserializer=repository__pb2.GetInfoAttributesRequest.FromString,
                    response_serializer=repository__pb2.GetInfoAttributesResponse.SerializeToString,
            ),
            'CalculateChecksum': grpc.unary_unary_rpc_method_handler(
                    servicer.CalculateChecksum,
                    request_deserializer=repository__pb2.CalculateChecksumRequest.FromString,
                    response_serializer=repository__pb2.CalculateChecksumResponse.SerializeToString,
            ),
            'GetSnapshot': grpc.unary_stream_rpc_method_handler(
                    servicer.GetSnapshot,
                    request_deserializer=repository__pb2.GetSnapshotRequest.FromString,
                    response_serializer=repository__pb2.GetSnapshotResponse.SerializeToString,
            ),
            'CreateRepositoryFromSnapshot': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateRepositoryFromSnapshot,
                    request_deserializer=repository__pb2.CreateRepositoryFromSnapshotRequest.FromString,
                    response_serializer=repository__pb2.CreateRepositoryFromSnapshotResponse.SerializeToString,
            ),
            'GetRawChanges': grpc.unary_stream_rpc_method_handler(
                    servicer.GetRawChanges,
                    request_deserializer=repository__pb2.GetRawChangesRequest.FromString,
                    response_serializer=repository__pb2.GetRawChangesResponse.SerializeToString,
            ),
            'SearchFilesByContent': grpc.unary_stream_rpc_method_handler(
                    servicer.SearchFilesByContent,
                    request_deserializer=repository__pb2.SearchFilesByContentRequest.FromString,
                    response_serializer=repository__pb2.SearchFilesByContentResponse.SerializeToString,
            ),
            'SearchFilesByName': grpc.unary_stream_rpc_method_handler(
                    servicer.SearchFilesByName,
                    request_deserializer=repository__pb2.SearchFilesByNameRequest.FromString,
                    response_serializer=repository__pb2.SearchFilesByNameResponse.SerializeToString,
            ),
            'RestoreCustomHooks': grpc.stream_unary_rpc_method_handler(
                    servicer.RestoreCustomHooks,
                    request_deserializer=repository__pb2.RestoreCustomHooksRequest.FromString,
                    response_serializer=repository__pb2.RestoreCustomHooksResponse.SerializeToString,
            ),
            'SetCustomHooks': grpc.stream_unary_rpc_method_handler(
                    servicer.SetCustomHooks,
                    request_deserializer=repository__pb2.SetCustomHooksRequest.FromString,
                    response_serializer=repository__pb2.SetCustomHooksResponse.SerializeToString,
            ),
            'BackupCustomHooks': grpc.unary_stream_rpc_method_handler(
                    servicer.BackupCustomHooks,
                    request_deserializer=repository__pb2.BackupCustomHooksRequest.FromString,
                    response_serializer=repository__pb2.BackupCustomHooksResponse.SerializeToString,
            ),
            'GetCustomHooks': grpc.unary_stream_rpc_method_handler(
                    servicer.GetCustomHooks,
                    request_deserializer=repository__pb2.GetCustomHooksRequest.FromString,
                    response_serializer=repository__pb2.GetCustomHooksResponse.SerializeToString,
            ),
            'GetObjectDirectorySize': grpc.unary_unary_rpc_method_handler(
                    servicer.GetObjectDirectorySize,
                    request_deserializer=repository__pb2.GetObjectDirectorySizeRequest.FromString,
                    response_serializer=repository__pb2.GetObjectDirectorySizeResponse.SerializeToString,
            ),
            'RemoveRepository': grpc.unary_unary_rpc_method_handler(
                    servicer.RemoveRepository,
                    request_deserializer=repository__pb2.RemoveRepositoryRequest.FromString,
                    response_serializer=repository__pb2.RemoveRepositoryResponse.SerializeToString,
            ),
            'ReplicateRepository': grpc.unary_unary_rpc_method_handler(
                    servicer.ReplicateRepository,
                    request_deserializer=repository__pb2.ReplicateRepositoryRequest.FromString,
                    response_serializer=repository__pb2.ReplicateRepositoryResponse.SerializeToString,
            ),
            'OptimizeRepository': grpc.unary_unary_rpc_method_handler(
                    servicer.OptimizeRepository,
                    request_deserializer=repository__pb2.OptimizeRepositoryRequest.FromString,
                    response_serializer=repository__pb2.OptimizeRepositoryResponse.SerializeToString,
            ),
            'PruneUnreachableObjects': grpc.unary_unary_rpc_method_handler(
                    servicer.PruneUnreachableObjects,
                    request_deserializer=repository__pb2.PruneUnreachableObjectsRequest.FromString,
                    response_serializer=repository__pb2.PruneUnreachableObjectsResponse.SerializeToString,
            ),
            'BackupRepository': grpc.unary_unary_rpc_method_handler(
                    servicer.BackupRepository,
                    request_deserializer=repository__pb2.BackupRepositoryRequest.FromString,
                    response_serializer=repository__pb2.BackupRepositoryResponse.SerializeToString,
            ),
            'RestoreRepository': grpc.unary_unary_rpc_method_handler(
                    servicer.RestoreRepository,
                    request_deserializer=repository__pb2.RestoreRepositoryRequest.FromString,
                    response_serializer=repository__pb2.RestoreRepositoryResponse.SerializeToString,
            ),
            'GetFileAttributes': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFileAttributes,
                    request_deserializer=repository__pb2.GetFileAttributesRequest.FromString,
                    response_serializer=repository__pb2.GetFileAttributesResponse.SerializeToString,
            ),
            'FastExport': grpc.unary_stream_rpc_method_handler(
                    servicer.FastExport,
                    request_deserializer=repository__pb2.FastExportRequest.FromString,
                    response_serializer=repository__pb2.FastExportResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'gitaly.RepositoryService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class RepositoryService(object):
    """RepositoryService is a service providing RPCs accessing repositories as a whole.
    """

    @staticmethod
    def RepositoryExists(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/RepositoryExists',
            repository__pb2.RepositoryExistsRequest.SerializeToString,
            repository__pb2.RepositoryExistsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RepositorySize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/RepositorySize',
            repository__pb2.RepositorySizeRequest.SerializeToString,
            repository__pb2.RepositorySizeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RepositoryInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/RepositoryInfo',
            repository__pb2.RepositoryInfoRequest.SerializeToString,
            repository__pb2.RepositoryInfoResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ObjectsSize(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/gitaly.RepositoryService/ObjectsSize',
            repository__pb2.ObjectsSizeRequest.SerializeToString,
            repository__pb2.ObjectsSizeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ObjectFormat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/ObjectFormat',
            repository__pb2.ObjectFormatRequest.SerializeToString,
            repository__pb2.ObjectFormatResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ApplyGitattributes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/ApplyGitattributes',
            repository__pb2.ApplyGitattributesRequest.SerializeToString,
            repository__pb2.ApplyGitattributesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FetchRemote(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/FetchRemote',
            repository__pb2.FetchRemoteRequest.SerializeToString,
            repository__pb2.FetchRemoteResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateRepository(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/CreateRepository',
            repository__pb2.CreateRepositoryRequest.SerializeToString,
            repository__pb2.CreateRepositoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetArchive(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.RepositoryService/GetArchive',
            repository__pb2.GetArchiveRequest.SerializeToString,
            repository__pb2.GetArchiveResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def HasLocalBranches(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/HasLocalBranches',
            repository__pb2.HasLocalBranchesRequest.SerializeToString,
            repository__pb2.HasLocalBranchesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FetchSourceBranch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/FetchSourceBranch',
            repository__pb2.FetchSourceBranchRequest.SerializeToString,
            repository__pb2.FetchSourceBranchResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Fsck(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/Fsck',
            repository__pb2.FsckRequest.SerializeToString,
            repository__pb2.FsckResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def WriteRef(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/WriteRef',
            repository__pb2.WriteRefRequest.SerializeToString,
            repository__pb2.WriteRefResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FindMergeBase(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/FindMergeBase',
            repository__pb2.FindMergeBaseRequest.SerializeToString,
            repository__pb2.FindMergeBaseResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateFork(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/CreateFork',
            repository__pb2.CreateForkRequest.SerializeToString,
            repository__pb2.CreateForkResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateRepositoryFromURL(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/CreateRepositoryFromURL',
            repository__pb2.CreateRepositoryFromURLRequest.SerializeToString,
            repository__pb2.CreateRepositoryFromURLResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateBundle(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.RepositoryService/CreateBundle',
            repository__pb2.CreateBundleRequest.SerializeToString,
            repository__pb2.CreateBundleResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateBundleFromRefList(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/gitaly.RepositoryService/CreateBundleFromRefList',
            repository__pb2.CreateBundleFromRefListRequest.SerializeToString,
            repository__pb2.CreateBundleFromRefListResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GenerateBundleURI(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/GenerateBundleURI',
            repository__pb2.GenerateBundleURIRequest.SerializeToString,
            repository__pb2.GenerateBundleURIResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FetchBundle(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/gitaly.RepositoryService/FetchBundle',
            repository__pb2.FetchBundleRequest.SerializeToString,
            repository__pb2.FetchBundleResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateRepositoryFromBundle(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/gitaly.RepositoryService/CreateRepositoryFromBundle',
            repository__pb2.CreateRepositoryFromBundleRequest.SerializeToString,
            repository__pb2.CreateRepositoryFromBundleResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.RepositoryService/GetConfig',
            repository__pb2.GetConfigRequest.SerializeToString,
            repository__pb2.GetConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FindLicense(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/FindLicense',
            repository__pb2.FindLicenseRequest.SerializeToString,
            repository__pb2.FindLicenseResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetInfoAttributes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.RepositoryService/GetInfoAttributes',
            repository__pb2.GetInfoAttributesRequest.SerializeToString,
            repository__pb2.GetInfoAttributesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CalculateChecksum(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/CalculateChecksum',
            repository__pb2.CalculateChecksumRequest.SerializeToString,
            repository__pb2.CalculateChecksumResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetSnapshot(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.RepositoryService/GetSnapshot',
            repository__pb2.GetSnapshotRequest.SerializeToString,
            repository__pb2.GetSnapshotResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateRepositoryFromSnapshot(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/CreateRepositoryFromSnapshot',
            repository__pb2.CreateRepositoryFromSnapshotRequest.SerializeToString,
            repository__pb2.CreateRepositoryFromSnapshotResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetRawChanges(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.RepositoryService/GetRawChanges',
            repository__pb2.GetRawChangesRequest.SerializeToString,
            repository__pb2.GetRawChangesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SearchFilesByContent(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.RepositoryService/SearchFilesByContent',
            repository__pb2.SearchFilesByContentRequest.SerializeToString,
            repository__pb2.SearchFilesByContentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SearchFilesByName(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.RepositoryService/SearchFilesByName',
            repository__pb2.SearchFilesByNameRequest.SerializeToString,
            repository__pb2.SearchFilesByNameResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RestoreCustomHooks(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/gitaly.RepositoryService/RestoreCustomHooks',
            repository__pb2.RestoreCustomHooksRequest.SerializeToString,
            repository__pb2.RestoreCustomHooksResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetCustomHooks(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/gitaly.RepositoryService/SetCustomHooks',
            repository__pb2.SetCustomHooksRequest.SerializeToString,
            repository__pb2.SetCustomHooksResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def BackupCustomHooks(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.RepositoryService/BackupCustomHooks',
            repository__pb2.BackupCustomHooksRequest.SerializeToString,
            repository__pb2.BackupCustomHooksResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetCustomHooks(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.RepositoryService/GetCustomHooks',
            repository__pb2.GetCustomHooksRequest.SerializeToString,
            repository__pb2.GetCustomHooksResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetObjectDirectorySize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/GetObjectDirectorySize',
            repository__pb2.GetObjectDirectorySizeRequest.SerializeToString,
            repository__pb2.GetObjectDirectorySizeResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RemoveRepository(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/RemoveRepository',
            repository__pb2.RemoveRepositoryRequest.SerializeToString,
            repository__pb2.RemoveRepositoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ReplicateRepository(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/ReplicateRepository',
            repository__pb2.ReplicateRepositoryRequest.SerializeToString,
            repository__pb2.ReplicateRepositoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def OptimizeRepository(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/OptimizeRepository',
            repository__pb2.OptimizeRepositoryRequest.SerializeToString,
            repository__pb2.OptimizeRepositoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PruneUnreachableObjects(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/PruneUnreachableObjects',
            repository__pb2.PruneUnreachableObjectsRequest.SerializeToString,
            repository__pb2.PruneUnreachableObjectsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def BackupRepository(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/BackupRepository',
            repository__pb2.BackupRepositoryRequest.SerializeToString,
            repository__pb2.BackupRepositoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RestoreRepository(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/RestoreRepository',
            repository__pb2.RestoreRepositoryRequest.SerializeToString,
            repository__pb2.RestoreRepositoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetFileAttributes(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/gitaly.RepositoryService/GetFileAttributes',
            repository__pb2.GetFileAttributesRequest.SerializeToString,
            repository__pb2.GetFileAttributesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FastExport(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/gitaly.RepositoryService/FastExport',
            repository__pb2.FastExportRequest.SerializeToString,
            repository__pb2.FastExportResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mercurial-repository.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import lint_pb2 as lint__pb2
from . import shared_pb2 as shared__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1amercurial-repository.proto\x12\x07hgitaly\x1a\nlint.proto\x1a\x0cshared.proto\"Y\n\x11InitConfigRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x16\n\x0enamespace_path\x18\x02 \x01(\t\"\x14\n\x12InitConfigResponse\"\x95\x02\n\x15HeptapodConfigSection\x12!\n\x14\x61llow_multiple_heads\x18\x02 \x01(\x08H\x00\x88\x01\x01\x12\x1c\n\x0f\x61llow_bookmarks\x18\x03 \x01(\x08H\x01\x88\x01\x01\x12\x45\n\x0c\x61uto_publish\x18\x04 \x01(\x0e\x32*.hgitaly.HeptapodConfigSection.AutoPublishH\x02\x88\x01\x01\"6\n\x0b\x41utoPublish\x12\x11\n\rWITHOUT_TOPIC\x10\x00\x12\x0b\n\x07NOTHING\x10\x01\x12\x07\n\x03\x41LL\x10\x02\x42\x17\n\x15_allow_multiple_headsB\x12\n\x10_allow_bookmarksB\x0f\n\r_auto_publish\"V\n\x17GetManagedConfigRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\r\n\x05local\x18\x02 \x01(\x08\"]\n\x18GetManagedConfigResponse\x12\x0f\n\x07inherit\x18\x01 \x01(\x08\x12\x30\n\x08heptapod\x18\x02 \x01(\x0b\x32\x1e.hgitaly.HeptapodConfigSection\"\xc2\x01\n\x17SetManagedConfigRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x14\n\x07inherit\x18\x02 \x01(\x08H\x00\x88\x01\x01\x12\x30\n\x08heptapod\x18\x03 \x01(\x0b\x32\x1e.hgitaly.HeptapodConfigSection\x12\x14\n\x0cremove_items\x18\x04 \x03(\t\x12\x0f\n\x07\x62y_line\x18\x05 \x01(\tB\n\n\x08_inherit\"\x1a\n\x18SetManagedConfigResponse\"\x8d\x01\n\x14GetConfigItemRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12(\n\x07\x61s_type\x18\x02 \x01(\x0e\x32\x17.hgitaly.ConfigItemType\x12\x0f\n\x07section\x18\x03 \x01(\t\x12\x0c\n\x04name\x18\x04 \x01(\t\"H\n\x15GetConfigItemResponse\x12\x13\n\tas_string\x18\x01 \x01(\tH\x00\x12\x11\n\x07\x61s_bool\x18\x02 \x01(\x08H\x00\x42\x07\n\x05value\">\n\x0eRecoverRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\"%\n\x0fRecoverResponse\x12\x12\n\nnot_needed\x18\x01 \x01(\x08\"?\n\x0fOptimizeRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\"\x12\n\x10OptimizeResponse\"\x88\x01\n\x12ResetCachesRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12\x31\n\x06\x63\x61\x63hes\x18\x02 \x03(\x0e\x32!.hgitaly.ResetCachesRequest.Cache\"\x11\n\x05\x43\x61\x63he\x12\x08\n\x04TAGS\x10\x00\"\x15\n\x13ResetCachesResponse\"b\n\rMercurialPeer\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x0f\n\x07ssh_key\x18\x05 \x01(\t\x12\x17\n\x0fssh_known_hosts\x18\x06 \x01(\t\x12\x1a\n\x12ssh_remote_command\x18\x07 \x01(\x0c\"\xa7\x01\n\x0bPushRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12+\n\x0bremote_peer\x18\x02 \x01(\x0b\x32\x16.hgitaly.MercurialPeer\x12%\n\x1donly_gitlab_branches_matching\x18\x03 \x03(\x0c\x12\x16\n\x0einclude_drafts\x18\x04 \x01(\x08\"&\n\x0cPushResponse\x12\x16\n\x0enew_changesets\x18\x01 \x01(\x08\"\xb8\x01\n\x0bPullRequest\x12,\n\nrepository\x18\x01 \x01(\x0b\x32\x12.gitaly.RepositoryB\x04\x98\xc6,\x01\x12+\n\x0bremote_peer\x18\x02 \x01(\x0b\x32\x16.hgitaly.MercurialPeer\x12\x1a\n\x04user\x18\x04 \x01(\x0b\x32\x0c.gitaly.User\x12\x1b\n\x13mercurial_revisions\x18\x05 \x03(\x0cJ\x04\x08\x03\x10\x04R\x0fgitlab_branches\"&\n\x0cPullResponse\x12\x16\n\x0enew_changesets\x18\x01 \x01(\x08*&\n\x0e\x43onfigItemType\x12\n\n\x06STRING\x10\x00\x12\x08\n\x04\x42OOL\x10\x01\x32\xe0\x05\n\x1aMercurialRepositoryService\x12M\n\nInitConfig\x12\x1a.hgitaly.InitConfigRequest\x1a\x1b.hgitaly.InitConfigResponse\"\x06\xfa\x97(\x02\x08\x01\x12V\n\rGetConfigItem\x12\x1d.hgitaly.GetConfigItemRequest\x1a\x1e.hgitaly.GetConfigItemResponse\"\x06\xfa\x97(\x02\x08\x02\x12\x44\n\x07Recover\x12\x17.hgitaly.RecoverRequest\x1a\x18.hgitaly.RecoverResponse\"\x06\xfa\x97(\x02\x08\x02\x12G\n\x08Optimize\x12\x18.hgitaly.OptimizeRequest\x1a\x19.hgitaly.OptimizeResponse\"\x06\xfa\x97(\x02\x08\x01\x12P\n\x0bResetCaches\x12\x1b.hgitaly.ResetCachesRequest\x1a\x1c.hgitaly.ResetCachesResponse\"\x06\xfa\x97(\x02\x08\x01\x12_\n\x10GetManagedConfig\x12 .hgitaly.GetManagedConfigRequest\x1a!.hgitaly.GetManagedConfigResponse\"\x06\xfa\x97(\x02\x08\x02\x12_\n\x10SetManagedConfig\x12 .hgitaly.SetManagedConfigRequest\x1a!.hgitaly.SetManagedConfigResponse\"\x06\xfa\x97(\x02\x08\x01\x12;\n\x04Push\x12\x14.hgitaly.PushRequest\x1a\x15.hgitaly.PushResponse\"\x06\xfa\x97(\x02\x08\x01\x12;\n\x04Pull\x12\x14.hgitaly.PullRequest\x1a\x15.hgitaly.PullResponse\"\x06\xfa\x97(\x02\x08\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mercurial_repository_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _INITCONFIGREQUEST.fields_by_name['repository']._options = None
  _INITCONFIGREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _GETMANAGEDCONFIGREQUEST.fields_by_name['repository']._options = None
  _GETMANAGEDCONFIGREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _SETMANAGEDCONFIGREQUEST.fields_by_name['repository']._options = None
  _SETMANAGEDCONFIGREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _GETCONFIGITEMREQUEST.fields_by_name['repository']._options = None
  _GETCONFIGITEMREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _RECOVERREQUEST.fields_by_name['repository']._options = None
  _RECOVERREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _OPTIMIZEREQUEST.fields_by_name['repository']._options = None
  _OPTIMIZEREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _RESETCACHESREQUEST.fields_by_name['repository']._options = None
  _RESETCACHESREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _PUSHREQUEST.fields_by_name['repository']._options = None
  _PUSHREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _PULLREQUEST.fields_by_name['repository']._options = None
  _PULLREQUEST.fields_by_name['repository']._serialized_options = b'\230\306,\001'
  _MERCURIALREPOSITORYSERVICE.methods_by_name['InitConfig']._options = None
  _MERCURIALREPOSITORYSERVICE.methods_by_name['InitConfig']._serialized_options = b'\372\227(\002\010\001'
  _MERCURIALREPOSITORYSERVICE.methods_by_name['GetConfigItem']._options = None
  _MERCURIALREPOSITORYSERVICE.methods_by_name['GetConfigItem']._serialized_options = b'\372\227(\002\010\002'
  _MERCURIALREPOSITORYSERVICE.methods_by_name['Recover']._options = None
  _MERCURIALREPOSITORYSERVICE.methods_by_name['Recover']._serialized_options = b'\372\227(\002\010\002'
  _MERCURIALREPOSITORYSERVICE.methods_by_name['Optimize']._options = None
  _MERCURIALREPOSITORYSERVICE.methods_by_name['Optimize']._serialized_options = b'\372\227(\002\010\001'
  _MERCURIALREPOSITORYSERVICE.methods_by_name['ResetCaches']._options = None
  _MERCURIALREPOSITORYSERVICE.methods_by_name['ResetCaches']._serialized_options = b'\372\227(\002\010\001'
  _MERCURIALREPOSITORYSERVICE.methods_by_name['GetManagedConfig']._options = None
  _MERCURIALREPOSITORYSERVICE.methods_by_name['GetManagedConfig']._serialized_options = b'\372\227(\002\010\002'
  _MERCURIALREPOSITORYSERVICE.methods_by_name['SetManagedConfig']._options = None
  _MERCURIALREPOSITORYSERVICE.methods_by_name['SetManagedConfig']._serialized_options = b'\372\227(\002\010\001'
  _MERCURIALREPOSITORYSERVICE.methods_by_name['Push']._options = None
  _MERCURIALREPOSITORYSERVICE.methods_by_name['Push']._serialized_options = b'\372\227(\002\010\001'
  _MERCURIALREPOSITORYSERVICE.methods_by_name['Pull']._options = None
  _MERCURIALREPOSITORYSERVICE.methods_by_name['Pull']._serialized_options = b'\372\227(\002\010\001'
  _globals['_CONFIGITEMTYPE']._serialized_start=1971
  _globals['_CONFIGITEMTYPE']._serialized_end=2009
  _globals['_INITCONFIGREQUEST']._serialized_start=65
  _globals['_INITCONFIGREQUEST']._serialized_end=154
  _globals['_INITCONFIGRESPONSE']._serialized_start=156
  _globals['_INITCONFIGRESPONSE']._serialized_end=176
  _globals['_HEPTAPODCONFIGSECTION']._serialized_start=179
  _globals['_HEPTAPODCONFIGSECTION']._serialized_end=456
  _globals['_HEPTAPODCONFIGSECTION_AUTOPUBLISH']._serialized_start=340
  _globals['_HEPTAPODCONFIGSECTION_AUTOPUBLISH']._serialized_end=394
  _globals['_GETMANAGEDCONFIGREQUEST']._serialized_start=458
  _globals['_GETMANAGEDCONFIGREQUEST']._serialized_end=544
  _globals['_GETMANAGEDCONFIGRESPONSE']._serialized_start=546
  _globals['_GETMANAGEDCONFIGRESPONSE']._serialized_end=639
  _globals['_SETMANAGEDCONFIGREQUEST']._serialized_start=642
  _globals['_SETMANAGEDCONFIGREQUEST']._serialized_end=836
  _globals['_SETMANAGEDCONFIGRESPONSE']._serialized_start=838
  _globals['_SETMANAGEDCONFIGRESPONSE']._serialized_end=864
  _globals['_GETCONFIGITEMREQUEST']._serialized_start=867
  _globals['_GETCONFIGITEMREQUEST']._serialized_end=1008
  _globals['_GETCONFIGITEMRESPONSE']._serialized_start=1010
  _globals['_GETCONFIGITEMRESPONSE']._serialized_end=1082
  _globals['_RECOVERREQUEST']._serialized_start=1084
  _globals['_RECOVERREQUEST']._serialized_end=1146
  _globals['_RECOVERRESPONSE']._serialized_start=1148
  _globals['_RECOVERRESPONSE']._serialized_end=1185
  _globals['_OPTIMIZEREQUEST']._serialized_start=1187
  _globals['_OPTIMIZEREQUEST']._serialized_end=1250
  _globals['_OPTIMIZERESPONSE']._serialized_start=1252
  _globals['_OPTIMIZERESPONSE']._serialized_end=1270
  _globals['_RESETCACHESREQUEST']._serialized_start=1273
  _globals['_RESETCACHESREQUEST']._serialized_end=1409
  _globals['_RESETCACHESREQUEST_CACHE']._serialized_start=1392
  _globals['_RESETCACHESREQUEST_CACHE']._serialized_end=1409
  _globals['_RESETCACHESRESPONSE']._serialized_start=1411
  _globals['_RESETCACHESRESPONSE']._serialized_end=1432
  _globals['_MERCURIALPEER']._serialized_start=1434
  _globals['_MERCURIALPEER']._serialized_end=1532
  _globals['_PUSHREQUEST']._serialized_start=1535
  _globals['_PUSHREQUEST']._serialized_end=1702
  _globals['_PUSHRESPONSE']._serialized_start=1704
  _globals['_PUSHRESPONSE']._serialized_end=1742
  _globals['_PULLREQUEST']._serialized_start=1745
  _globals['_PULLREQUEST']._serialized_end=1929
  _globals['_PULLRESPONSE']._serialized_start=1931
  _globals['_PULLRESPONSE']._serialized_end=1969
  _globals['_MERCURIALREPOSITORYSERVICE']._serialized_start=2012
  _globals['_MERCURIALREPOSITORYSERVICE']._serialized_end=2748
# @@protoc_insertion_point(module_scope)



# Copyright 2024 Georges Racinet <georges.racinet@cloudcrane.io>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import re

from mercurial import (
    error as hgerror,
)

from .errors import (
    MercurialPathNotFound,
    MercurialChangesetNotFound,
)
from .oid import extract_blob_oid
from .stub.analysis_pb2 import (
    CheckBlobsGeneratedRequest,
)


def frozen_bytes_set(it):
    return frozenset(s.encode('ascii') for s in it)


# compare go-enry/data/generated.go
GENERATED_FILE_PATH_REGEXPS = [
    re.compile(b'|'.join((
        rb'(^Pods|/Pods)/',  # Cocoa pods
        rb'(^|/)Carthage/Build/',  # Carthage build
        rb'\.feature\.css$',  # Generated NET specflow feature file
        rb'node_modules/',  # Node modules
        # Go vendor
        rb'vendor/([-0-9A-Za-z]+\.)+(com|edu|gov|in|me|net|org|fm|io)',
        rb'(Gopkg|glide).lock$',  # Go lock
        rb'(^|/)(\w+\.)?esy.lock$`',  # Esy lock
        rb'npm-shrinkwrap.json$',
        rb'package-lock.json$',  # NPM package lock
        rb'(^|/)\.pnp\.(c|m)?js$',  # Yarn plugnplay
        rb'Godeps/',
        rb'composer.lock$',
        rb'.\.zep\.(?:c|h|php)$',  # Generated by zephir
        rb'Cargo.lock$',
        rb'Pipfile.lock$',  # Pipenv lock
        rb'__generated__/',  # GraphQL relay
        rb'\.(js|css)\.map$',  # Source map
    ))),
    re.compile(b'|'.join((
        rb'\.designer\.(cs|vb)$',  # NET designer file
    )), re.IGNORECASE),
]

# compare go-enry/data/generated.go
GENERATED_FILE_EXTENSIONS = frozen_bytes_set((
    # Xcode
    'nib',
    'xcworkspacedata',
    'xcuserstate',
    # TODO complete the list
))

GENERATED_LINE_RX = re.compile(
    b'|'.join((
        # Source Map
        rb'\A{"version":\d+',
        rb'\A/\*\*Begin line maps\. \*\*/{',
        # Generated NetDoc  TODO should be only for .xml files
        rb'\A.*$.*<doc>.*(\r)?\n.*<assembly>(\r)?\n.*</doc>(\r)?\n.*\Z',
        # PEG.js
        rb'^(?:[^/]|/[^\*])*/\*(?:[^\*]|\*[^/])*Generated by PEG.js',
        # Generated Go (actually our version is more general, but anything
        # matching this regexp is looking for trouble!
        rb'Code generated by',
        # Protocol Buffers, general case
        b"Generated by the protocol buffer compiler.  DO NOT EDIT!",
        # Protocol Buffers, JavaScript (should be only line 6 and crossed with
        # extension, but given the content, matching this is acceptable
        # TODO unless we are looking at a code generator!
        b"GENERATED CODE -- DO NOT EDIT!",
        # Apache Thrift
        b"Autogenerated by Thrift Compiler",
        # JNI header (again more general than go-enry, because we don't want
        # to assert it to be JNI, just that it is generated)
        rb"/\* DO NOT EDIT THIS FILE - it is machine generated \*/",
        # VCR Cassette TODO should be for .yml files only
        rb"\A.*(\r)?\n.*recorded_with: VCR",
        # Compiled Cython TODO should be .c files only
        rb"\A.*Generated by Cython",
        # (Fortran) module
        rb"\A.*(PCBNEW-LibModule-V|GFORTRAN module version ')",
        # Racc
        rb"# This file is automatically generated by Racc",
        # JFlex
        rb"/\* The following code was generated by JFlex ",
        # Grammar Kit
        rb"// This is a generated file. Not intended for manual editing\.",
        # Roxygen 2
        rb"% Generated by roxygen2: do not edit by hand",
        # Jison
        rb"/\* (parser generated by jison |generated by jison-lex )",
        # gRPC C++
        rb"// Generated by the gRPC",
        # Dart
        rb"generated code\W{2,3}do not modify",
        # Perl's PPPort
        rb"Automatically created by Devel::PPPort",
        # GIMP
        rb'/\* GIMP [a-zA-Z0-9\- ]+ C\-Source image dump \(.+?\.c\) \*/',
        (rb'/\*  GIMP header image file format '
         rb'\([a-zA-Z0-9\- ]+\)\: .+?\.h  \*/'),
        # Visual Studio 6
        rb'# Microsoft Developer Studio Generated Build File',
        # Haxe
        rb"Generated by Haxe",
        # Jooq
        rb"This file is generated by jOOQ\.",
    )),
    re.MULTILINE
)

# TODO isMinifiedFile, hasSourceMapReference, isCompiledCoffeeScript,
# isGeneratedPostScript,
# isGeneratedUnity3DMeta (extension restrictions should really apply)
# isGeneratedGameMakerStudio, isGeneratedHTML


def is_blob_generated(repo, blob: CheckBlobsGeneratedRequest.Blob):
    """"Detect if a blob is made of generated data."""
    # For now the Rails application sends revisions in `commit_id:path`
    # format, so close to HGitaly blob oids (!). Probably upstream does not
    # want to use actual Git blob oids, but Gitaly does support them
    colon_split = blob.revision.split(b':', 1)
    if len(colon_split) == 2:
        csid, rpath = colon_split
    else:  # direct oid
        csid, rpath = extract_blob_oid(repo, blob.revision.decode('utf-8'))

    if csid is None or rpath is None:
        raise ValueError("Invalid blob oid")

    extension = rpath.rsplit(b'.', 1)[-1]
    if extension in GENERATED_FILE_EXTENSIONS:
        return True

    if any(rx.search(rpath) is not None
           for rx in GENERATED_FILE_PATH_REGEXPS):
        return True

    try:
        changeset = repo[csid]
    except hgerror.RepoLookupError:
        raise MercurialChangesetNotFound(csid)

    try:
        filectx = changeset.filectx(rpath)
    except KeyError:
        raise MercurialPathNotFound(rpath)

    return GENERATED_LINE_RX.search(filectx.data()) is not None



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import logging
from grpc import StatusCode

from mercurial import (
    error as hgerror,
)

from heptapod.gitlab.branch import (
    GITLAB_BRANCH_REF_PREFIX,
    gitlab_branch_from_ref,
    gitlab_branch_ref,
)
from heptapod.gitlab.tag import (
    GITLAB_TAG_REF_PREFIX,
)
from hgext3rd.heptapod.branch import get_default_gitlab_branch
from hgext3rd.heptapod.typed_ref import GITLAB_TYPED_REFS_MISSING
from hgext3rd.heptapod.special_ref import (
    parse_special_ref,
    special_refs,
    write_special_refs,
)
from hgext3rd.heptapod.keep_around import (
    parse_keep_around_ref,
)

from ..errors import (
    not_implemented,
    structured_abort,
)
from ..logging import LoggerAdapter
from ..pagination import (
    extract_limit,
)
from ..revision import (
    CHANGESET_HASH_BYTES_REGEXP,
    gitlab_revision_hash,
)
from ..stream import (
    chunked_limit,
)
from ..stub.shared_pb2 import (
    Branch,
    SortDirection,
)
from ..stub.errors_pb2 import (
    ReferenceNotFoundError,
)
from ..stub.ref_pb2 import (
    FindDefaultBranchNameRequest,
    FindDefaultBranchNameResponse,
    FindLocalBranchesRequest,
    FindLocalBranchesResponse,
    FindAllBranchesRequest,
    FindAllBranchesResponse,
    FindAllTagsRequest,
    FindAllTagsResponse,
    FindRefsByOIDRequest,
    FindRefsByOIDResponse,
    FindTagError,
    FindTagRequest,
    FindTagResponse,
    FindAllRemoteBranchesRequest,
    FindAllRemoteBranchesResponse,
    RefExistsRequest,
    RefExistsResponse,
    FindBranchRequest,
    FindBranchResponse,
    DeleteRefsRequest,
    DeleteRefsResponse,
    ListBranchNamesContainingCommitRequest,
    ListBranchNamesContainingCommitResponse,
    ListTagNamesContainingCommitRequest,
    ListTagNamesContainingCommitResponse,
    GetTagSignaturesRequest,
    GetTagSignaturesResponse,
    GetTagMessagesRequest,
    GetTagMessagesResponse,
    ListRefsRequest,
    ListRefsResponse,
)
from ..stub.ref_pb2_grpc import RefServiceServicer

from ..revision import (
    ZERO_SHA_STR,
)
from ..branch import (
    BranchSortBy,
    gitlab_branch_head,
    iter_gitlab_branches,
    iter_gitlab_branches_as_refs,
    sorted_gitlab_branches_as_refs,
)
from ..gitlab_ref import (
    ensure_special_refs,
    has_keep_around,
    iter_gitlab_special_refs_as_refs,
    iter_keep_arounds_as_refs,
)
from ..tag import (
    EXCLUDED_TAG_TYPES,
    iter_gitlab_tags,
    iter_gitlab_tags_as_refs,
)
from .. import message
from ..servicer import HGitalyServicer
from ..util import chunked

base_logger = logging.getLogger(__name__)
DEFAULT_BRANCH_FILE_NAME = b'default_gitlab_branch'
FIND_LOCAL_BRANCHES_SORT_BY_TO_INNER = {
    FindLocalBranchesRequest.SortBy.NAME: BranchSortBy.FULL_REF_NAME,
    FindLocalBranchesRequest.SortBy.UPDATED_ASC: BranchSortBy.UPDATED_ASC,
    FindLocalBranchesRequest.SortBy.UPDATED_DESC: BranchSortBy.UPDATED_DESC
}

FIND_REFS_BY_OID_SORT_FIELD_TO_LIST_REFS = dict(
    refname=ListRefsRequest.SortBy.REFNAME,
    creatordate=ListRefsRequest.SortBy.CREATORDATE,
    authordate=ListRefsRequest.SortBy.AUTHORDATE,
)
"""keys are amenable fields in git-for-each-ref(1)

Not many such keys actually map to a SortBy value for ListRefsRequest.
"""


class RefServicer(RefServiceServicer, HGitalyServicer):
    """RefService implementation.

    The ordering of methods in this source file is the same as in the proto
    file.
    """
    def FindDefaultBranchName(
            self,
            request: FindDefaultBranchNameRequest,
            context) -> FindDefaultBranchNameResponse:
        logger = LoggerAdapter(base_logger, context)
        repo = self.load_repo(request.repository, context)

        branch = get_default_gitlab_branch(repo)
        if branch is None or gitlab_branch_head(repo, branch) is None:
            # Very coincidental, but this ends up as empty GitLab branch name
            # which the chain of || in app/models/concerns/repository
            # that eventually turns this in `nil`, which is the marker
            # used in the hg_fixup_default_branch in PostReceive…
            # TODO now that we have richer notifications, make a clear one
            # for that case
            logger.warning("FindDefaultBranchName: no information stored "
                           "for repo at %r, returning hempty response",
                           repo)
            return FindDefaultBranchNameResponse()

        return FindDefaultBranchNameResponse(name=gitlab_branch_ref(branch))

    def FindLocalBranches(self,
                          request: FindLocalBranchesRequest,
                          context) -> FindLocalBranchesResponse:
        repo = self.load_repo(request.repository, context)
        limit = extract_limit(request)
        if limit == 0:
            # util.chunk consider this means no limit,
            # shared.proto defines it to mean an empty response
            return

        after = request.pagination_params.page_token
        # TODO encoding?
        after = after.encode('utf-8') if after else None
        sort_by = FIND_LOCAL_BRANCHES_SORT_BY_TO_INNER.get(request.sort_by)

        for chunk in chunked(sorted_gitlab_branches_as_refs(repo,
                                                            sort_by=sort_by,
                                                            after=after),
                             limit=limit):
            yield FindLocalBranchesResponse(
                local_branches=(message.branch(name, head)
                                for name, head in chunk)
            )

    def FindAllBranches(self,
                        request: FindAllBranchesRequest,
                        context) -> FindAllBranchesResponse:
        Branch = FindAllBranchesResponse.Branch
        repo = self.load_repo(request.repository, context)
        for chunk in chunked(iter_gitlab_branches(repo)):
            yield FindAllBranchesResponse(
                branches=(Branch(name=name, target=message.commit(head))
                          for name, head in chunk))

    def FindAllTags(self,
                    request: FindAllTagsRequest,
                    context) -> FindAllTagsResponse:
        repo = self.load_repo(request.repository, context)
        for chunk in chunked(iter_gitlab_tags(repo)):
            yield FindAllTagsResponse(tags=[message.tag(name, ctx)
                                            for name, ctx in chunk])

    def FindTag(self,
                request: FindTagRequest,
                context) -> FindTagResponse:
        repo = self.load_repo(request.repository, context)
        name = request.tag_name
        if repo.tagtype(name) in EXCLUDED_TAG_TYPES:
            structured_abort(
                context, StatusCode.NOT_FOUND, "tag does not exist",
                FindTagError(tag_not_found=ReferenceNotFoundError(
                    reference_name=b'refs/tags/' + name)))

        node = repo.tags()[name]
        return FindTagResponse(tag=message.tag(name, repo[node]))

    def FindAllRemoteBranches(self,
                              request: FindAllRemoteBranchesRequest,
                              context) -> FindAllRemoteBranchesResponse:
        """There is no concept of "remote branch" in Mercurial."""
        return iter(())

    def RefExists(self,
                  request: RefExistsRequest,
                  context) -> RefExistsResponse:
        ref = request.ref
        if not ref.startswith(b'refs/'):
            context.abort(StatusCode.INVALID_ARGUMENT, "invalid refname")

        # TODO protect here
        repo = self.load_repo(request.repository, context)

        gl_branch = gitlab_branch_from_ref(ref)
        if gl_branch is not None:
            return RefExistsResponse(
                value=gitlab_branch_head(repo, gl_branch) is not None)

        gl_tag = gitlab_tag_from_ref(ref)
        if gl_tag is not None:
            return RefExistsResponse(
                value=repo.tagtype(gl_tag) not in EXCLUDED_TAG_TYPES)

        special = parse_special_ref(ref)
        if special is not None:
            srefs = special_refs(repo)
            if srefs is GITLAB_TYPED_REFS_MISSING:
                srefs = ensure_special_refs(repo)

            return RefExistsResponse(value=special in srefs)

        keep_around = parse_keep_around_ref(ref)
        if keep_around is not None:
            return RefExistsResponse(value=has_keep_around(repo, keep_around))
        return RefExistsResponse(value=False)

    def FindBranch(self,
                   request: FindBranchRequest,
                   context) -> FindBranchResponse:
        repo = self.load_repo(request.repository, context)
        name = request.name
        if name.startswith(b'refs/'):
            name = gitlab_branch_from_ref(request.name)

        if name is None:
            # TODO SPEC check if we really must exclude other refs
            return FindBranchResponse(branch=None)

        head = gitlab_branch_head(repo, name)
        if head is None:
            return FindBranchResponse(branch=None)

        return FindBranchResponse(
            branch=Branch(name=name, target_commit=message.commit(head)))

    def DeleteRefs(self,
                   request: DeleteRefsRequest,
                   context) -> DeleteRefsResponse:
        except_prefix = request.except_with_prefix
        refs = request.refs
        if refs and except_prefix:
            context.abort(StatusCode.INVALID_ARGUMENT,
                          "ExceptWithPrefix and Refs are mutually exclusive")

        repo = self.load_repo(request.repository, context)
        srefs = special_refs(repo)
        if srefs is GITLAB_TYPED_REFS_MISSING:
            srefs = ensure_special_refs(repo)
        if refs:
            # Using copy() to avoid doing anything (natural rollback) if
            # one of the ref is bogus.
            # It's not really important right now because we have
            # no cache of loaded repos, but that will change sooner or later.
            srefs = srefs.copy()
            err = remove_special_refs(srefs, refs)
            if err is not None:
                return DeleteRefsResponse(git_error=err)

        if except_prefix:
            srefs = special_refs_matching_prefixes(srefs, except_prefix)

        write_special_refs(repo, srefs)
        return DeleteRefsResponse()

    def ListBranchNamesContainingCommit(
            self,
            request: ListBranchNamesContainingCommitRequest,
            context) -> ListBranchNamesContainingCommitResponse:
        repo = self.load_repo(request.repository, context)

        gl_branches_by_heads = {}
        for gl_branch, head in iter_gitlab_branches(repo):
            rev = head.rev()
            gl_branches_by_heads.setdefault(rev, []).append(gl_branch)

        repo = repo.unfiltered()
        heads = repo.revs(b'%ld and %s::', gl_branches_by_heads,
                          request.commit_id)
        # TODO SPEC since there's a limit, we'll have to know what is
        # the expected ordering.
        #
        # In Gitaly sources, this is in refnames.go, which in turns call
        # `git for-each-ref` without a `sort` option. Then according to
        # the man page:
        #  --sort=<key>
        #      A field name to sort on. Prefix - to sort in descending order
        #      of the value. When unspecified, refname is used.
        for chunk in chunked((gl_branch
                              for head in heads
                              for gl_branch in gl_branches_by_heads[head]),
                             limit=request.limit):
            yield ListBranchNamesContainingCommitResponse(
                branch_names=chunk)

    def ListTagNamesContainingCommit(
            self,
            request: ListTagNamesContainingCommitRequest,
            context) -> ListTagNamesContainingCommitResponse:
        # TODO support ordering, see similar method for branches
        repo = self.load_repo(request.repository, context)

        try:
            revs = repo.revs("%s:: and tag()", request.commit_id)
        except hgerror.FilteredRepoLookupError:
            revs = []

        tag_names = (name for rev in revs for name in repo[rev].tags()
                     if repo.tagtype(name) not in EXCLUDED_TAG_TYPES)
        for chunk in chunked(tag_names, limit=request.limit):
            yield ListTagNamesContainingCommitResponse(tag_names=chunk)

    def GetTagSignatures(self,
                         request: GetTagSignaturesRequest,
                         context) -> GetTagSignaturesResponse:
        not_implemented(context, issue=75)  # pragma no cover

    def GetTagMessages(self,
                       request: GetTagMessagesRequest,
                       context) -> GetTagMessagesResponse:
        """Return messages of the given tags.

        In Mercurial, all tags have messages, and these are descriptions
        of the changests that give them values.

        For now, we'll consider that the id of a tag is the nod id of the
        changeset that gives it its current value.
        """
        repo = self.load_repo(request.repository, context)
        # TODO check that the given id is indeed for a tag, i.e. a
        # changeset that affects .hgtags?
        for tag_id in request.tag_ids:
            yield GetTagMessagesResponse(tag_id=tag_id,
                                         message=repo[tag_id].description())

    def ListNewCommits(self, request, context):
        """Not relevant for Mercurial

        From ``ref.proto``:
            Returns commits that are only reachable from the ref passed

        But actually, the request has a ``commit_id`` field, and it's not
        ``bytes``, hence can't be used for a ref.

        The reference Gitaly implementation is in `list_new_commits.go`.
        It boils down to::

          git rev-list --not --all ^oid

        with ``oid`` being ``request.commit id`` (not really a ref, then).
        additional comment: "the added ^ is to negate the oid since there is
        a --not option that comes earlier in the arg list"

        Note that ``--all`` includes Git refs that are descendents of the
        given commit. In other words, the results are ancestors of the
        given commit that would be garbage collected unless they get a ref
        soon.

        In the whole of the GitLab FOSS code base (as of GitLab 12.10),
        this is used only in pre-receive changes checks, i.e, before any ref
        has been assigned to commits that are new, indeed.

        We'll need a Mercurial specific version of the pre-receive check
        anyway.

        With HGitaly, all Mercurial changesets are at least ancestors of
        a GitLab branch head, the only exception not being closed heads,
        which are not mapped, so the results should boil down to something like
        ``reverse(::x and ::(heads() and closed()))``
        """
        raise NotImplementedError(
            "Not relevant for Mercurial")  # pragma: no cover

    def ListNewBlobs(self, request, context):
        """Not relevant for Mercurial.

        This is the same as :meth:`ListNewCommits()`, returning the blobs.
        In Gitaly sources, this is done by adding ``--objects`` to the
        otherwise same call to ``git rev-list`` as for :meth:`ListNewCommits`.

        As of GitLab 12.10, this is used only in
        ``GitAccess#check_changes_size`` (enforcing size limits for pushes).
        """
        raise NotImplementedError(
            "Not relevant for Mercurial")  # pragma: no cover

    def iter_refs(self, repo, request: ListRefsRequest, context):
        """"Generator yielding refernece names and target shas.

        For very direct use in ListRefs and similar.
        """
        patterns = request.patterns

        refs = []
        if patterns == [b'refs/'] and len(repo):
            refs.append((b'ALL', ZERO_SHA_STR))
        if request.head:
            branch = get_default_gitlab_branch(repo)
            if branch is not None:
                head = gitlab_branch_head(repo, branch)
                if head is not None:
                    refs.append((b'HEAD', head.hex().decode('ascii')))
        if patterns == [GITLAB_BRANCH_REF_PREFIX]:
            refs.extend(iter_gitlab_branches_as_refs(repo, deref=False))
        elif patterns == [GITLAB_TAG_REF_PREFIX]:
            refs.extend(iter_gitlab_tags_as_refs(repo, deref=False))
        else:
            if patterns:
                patterns = tuple(pat.split(b'/') for pat in patterns)
            refs.extend(iter_gitlab_branches_as_refs(repo,
                                                     deref=False,
                                                     patterns=patterns))
            refs.extend(iter_gitlab_tags_as_refs(repo,
                                                 deref=False,
                                                 patterns=patterns))
            refs.extend(iter_gitlab_special_refs_as_refs(repo,
                                                         deref=False,
                                                         patterns=patterns))
            refs.extend(iter_keep_arounds_as_refs(repo,
                                                  deref=False,
                                                  patterns=patterns))

        pointing_at = request.pointing_at_oids
        if pointing_at:
            refs = [r for r in refs
                    if r[1].encode('ascii') in pointing_at or r[0] == b'HEAD']

        # ignore `peel_tags`:
        # Mercurial has no real equivalent of Git's annotated tags. The
        # closest notion would be the tagging changeset, but that is not
        # well-defined. Hence Mercurial tags always point to the target commit.
        # We could force the `peeled_target` attribute of the `Reference`
        # messages, but that would not bring us to better compliance: the
        # protocol comment states explicitely that it should stay empty if
        # the object is not an annotated tag, and this is confirmed by
        # Gitaly Comparison tests (tag created by hg-git in this case).
        # On the contrary, filling-in `peeled_target` could very well make
        # clients believe that the tag is annotated, leading to dedicated
        # actions that can only go wrong.

        if len(refs) < (2 if request.head else 1):
            # avoid yielding an empty response, as Gitaly does not.
            # TODO consider doing that in the `chunked()` helper,
            # this looks to be rather systematic.
            # also in this case, even if `head` is `True`, in the Gitaly
            # impl, it looks like nothing being read from Git does not
            # trigger any sending, even though the sender created by
            # `newListRefsWriter` would emit HEAD first in any case… if it
            # emitted anything. See Gitaly Comparison test for validation of
            # this.
            return

        sort_by = request.sort_by
        if sort_by.key == ListRefsRequest.SortBy.Key.REFNAME:
            sort_key_fun = None
        else:
            # Some references can point to obsolete changesets
            # (e.g, keep-arounds). In any case it is certainly not the
            # role of this method to prevent one to do so.
            unfi = repo.unfiltered()

            # all other choices are about dates, and we have no other choice
            # that conflating them with the commit date of the ref target
            def sort_key_fun(ref_tuple):
                sha = ref_tuple[1]
                return unfi[sha].date()

        refs.sort(key=sort_key_fun,
                  reverse=sort_by.direction == SortDirection.DESCENDING)

        for chunk in chunked(refs):
            yield chunk

    def ListRefs(self, request: ListRefsRequest,
                 context) -> ListRefsResponse:
        repo = self.load_repo(request.repository, context)
        Reference = ListRefsResponse.Reference
        for chunk in self.iter_refs(repo, request, context):
            yield ListRefsResponse(
                references=(Reference(name=name, target=target)
                            for name, target in chunk))

    def FindRefsByOID(self, request: FindRefsByOIDRequest,
                      context) -> FindRefsByOIDResponse:
        repo = self.load_repo(request.repository, context)
        oid = request.oid.encode('utf8')
        if not CHANGESET_HASH_BYTES_REGEXP.match(oid):
            # a bit wider than needed, would probably be more
            # efficient to query nodemap directly, but we'll consider this
            # the day we reimplement in RHGitaly
            oid = gitlab_revision_hash(repo, oid)
        sort_by = FIND_REFS_BY_OID_SORT_FIELD_TO_LIST_REFS.get(
            request.sort_field, ListRefsRequest.SortBy.REFNAME)

        patterns = request.ref_patterns
        if patterns:
            patterns = [pat.encode('utf8') for pat in patterns]
        else:
            patterns = [b'refs/heads/', b'refs/tags/']

        limit = request.limit
        if limit == 0:  # said to mean no limit in protocol comment
            limit = None

        refs = []
        for chunk in chunked_limit(self.iter_refs(
                repo,
                ListRefsRequest(
                    repository=request.repository,
                    pointing_at_oids=[oid],
                    patterns=patterns,
                    sort_by=ListRefsRequest.SortBy(key=sort_by),
                ),
                context), limit):
            refs.extend(name for name, _tgt in chunk)
        return FindRefsByOIDResponse(refs=refs)


def gitlab_tag_from_ref(ref):
    if ref.startswith(GITLAB_TAG_REF_PREFIX):
        return ref[len(GITLAB_TAG_REF_PREFIX):]


def remove_special_refs(special_refs, to_remove):
    """Remove given elements of the ``special_refs`` mapping.

    :param dict special_refs: the special refs to remove from, whose keys
       are the shortened special ref name.
    :param to_remove: iterable of full ref names to remove.
    :returns: ``None`` if ok, else an error message.

    It is not an error to remove an absent ref, but it is one to request
    removal of a ref that is not a special ref.
    """
    for ref in to_remove:
        name = parse_special_ref(ref)
        if name is None:
            return (
                "Only special refs, such as merge-requests (but "
                "not keep-arounds) can be directly deleted in Mercurial, "
                "got %r" % ref)
        special_refs.pop(name, None)


def special_refs_matching_prefixes(special_refs, prefixes):
    """Return a sub-mapping of special refs matching given prefixes

    :param prefixes: given as full ref names, e.g. `refs/pipelines`.
        Any such prefix that would not be a special ref is simply ignored.
    """
    # Don't use parse_special_ref here, even if tempting: it would
    # not accept prefixes without trailing slashes, such as `refs/pipelines`
    # nor more partial prefixes, such as `refs/pipe`. Currently all used
    # prefixes are full with trailing slash, but if upstream developers
    # ever use a more partial prefix, this would result in data loss
    short_prefixes = {pref[5:] for pref in prefixes
                      if pref.startswith(b'refs/')}

    return {name: target for name, target in special_refs.items()
            if any(name.startswith(prefix) for prefix in short_prefixes)}



import os
import psutil

IS_CHILD_PROCESS = False


def is_current_service_process(pid):
    """Determine whether the given pid is part of the current HGitaly service.

    For now, the logic is that other processes from the same HGitaly service
    are expected to be siblings of the current process, unless in the special
    case where HGitaly is not started as a prefork server, which should happen
    with the `--mono-process` option (debuging sessions) and tests only.
    """
    this_pid = os.getpid()

    if not IS_CHILD_PROCESS:
        return pid == this_pid

    try:
        proc = psutil.Process(pid)
    except psutil.NoSuchProcess:
        return False

    return proc.ppid() == os.getppid()



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import pytest

from mercurial.node import nullrev

from heptapod.testhelpers import (
    LocalRepoWrapper,
)
from hgext3rd.heptapod.special_ref import (
    write_gitlab_special_ref,
)
from hgext3rd.heptapod.keep_around import (
    create_keep_around,
)

from ..gitlab_ref import (
    keep_around_ref_path,
)
from ..revision import (
    ALL_CHANGESETS,
    RevisionNotFound,
    gitlab_revision_changeset,
    gitlab_revision_hash,
    resolve_revspecs_positive_negative,
)


def make_repo(path):
    return LocalRepoWrapper.init(path,
                                 config=dict(
                                     extensions=dict(topic='', evolve=''),
                                 ))


def test_gitlab_revision_changeset_by_hex(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo

    ctx = wrapper.write_commit('foo')

    assert gitlab_revision_changeset(repo, ctx.hex()) == ctx

    # special node ids
    # collision very unlikely, but still a prefix
    assert gitlab_revision_changeset(repo, b'f' * 39) is None
    assert gitlab_revision_changeset(repo, b'0' * 39).rev() == nullrev
    # full node hex
    assert gitlab_revision_changeset(repo, b'f' * 40) is None
    assert gitlab_revision_changeset(repo, b'00000000' * 5).rev() == nullrev

    wrapper.command('amend', message=b'amended')

    obs_ctx = gitlab_revision_changeset(repo, ctx.hex())
    assert obs_ctx == ctx
    assert obs_ctx.obsolete()


def test_gitlab_revision_changeset_empty_repo(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo

    assert gitlab_revision_changeset(repo, b'HEAD') is None


def test_gitlab_revision_special_ref(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo

    ctx = wrapper.write_commit('foo')
    ref_name = b'merge-requests/1/head'
    ref_path = b'refs/merge-requests/1/head'

    write_gitlab_special_ref(repo, ref_name, ctx)
    assert gitlab_revision_changeset(repo, ref_path) == ctx


def test_gitlab_revision_keep_around(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo

    ctx = wrapper.write_commit('foo')
    sha = ctx.hex()
    create_keep_around(repo, sha)

    assert gitlab_revision_changeset(repo, keep_around_ref_path(sha)) == ctx
    assert gitlab_revision_changeset(
        repo, keep_around_ref_path(b'cafe' * 10)) is None


def test_gitlab_revision_gl_branch(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo
    ctx = wrapper.write_commit('foo')

    assert (
        gitlab_revision_changeset(repo, b'refs/heads/branch/default')
        == ctx
    )
    assert gitlab_revision_changeset(repo, b'branch/default') == ctx

    # precise ref form can be for nothing but a branch
    # here, just stripping the prefix would end over to direct lookup by
    # tag, bookmark or node ID
    assert gitlab_revision_changeset(repo, b'refs/heads/' + ctx.hex()) is None


def test_gitlab_revision_node_id(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo
    ctx = wrapper.write_commit('foo')

    assert gitlab_revision_changeset(repo, ctx.hex()) == ctx
    assert gitlab_revision_changeset(repo, b'12ca34fe' * 5) is None


def test_gitlab_revision_hash(tmpdir):
    wrapper = make_repo(tmpdir)
    repo = wrapper.repo
    ctx = wrapper.write_commit('foo')

    assert gitlab_revision_hash(repo, b'branch/default') == ctx.hex()
    with pytest.raises(RevisionNotFound) as exc_info:
        gitlab_revision_hash(repo, b'unknown')
    assert exc_info.value.args == (b'unknown', )


def test_resolve_revspecs_positive_negative(tmpdir):
    wrapper = make_repo(tmpdir)
    sha1 = wrapper.write_commit('foo').hex()
    sha2 = wrapper.write_commit('bar', branch='other').hex()

    def resolve(revisions, **kw):
        return resolve_revspecs_positive_negative(wrapper.repo, revisions,
                                                  **kw)

    empty = set()
    assert resolve([b'branch/default']) == ({sha1}, empty)
    assert resolve([b'branch/default',
                    b'branch/other']) == ({sha1, sha2}, empty)
    assert resolve([b'branch/default',
                    sha2,
                    b'^branch/other']) == ({sha1, sha2}, {sha2})
    assert resolve([b'^branch/default',
                    sha2,
                    b'branch/other']) == ({sha2}, {sha1})
    assert resolve([sha2,
                    b'^branch/default',
                    b'branch/other']) == ({sha2}, {sha1})

    # cases with the `ALL` pseudo ref (stops further resolution and supersedes
    # all previously done ones)
    assert resolve([b'ALL']) == (ALL_CHANGESETS, empty)
    assert resolve([b'ALL',
                    b'branch/other']) == (ALL_CHANGESETS, empty)
    assert resolve([b'branch/default',
                    b'ALL',
                    b'^branch/other']) == (ALL_CHANGESETS, {sha2})
    assert resolve([b'^branch/default',
                    b'ALL',
                    b'branch/other']) == (ALL_CHANGESETS, {sha1})
    assert resolve([b'ALL',
                    b'^branch/default',
                    b'branch/other']) == (ALL_CHANGESETS, {sha1})
    assert resolve([b'ALL',
                    b'unknown']) == (ALL_CHANGESETS, empty)

    # cases of raising because of unknown revision
    with pytest.raises(RevisionNotFound) as exc_info:
        resolve([b'unknown', b'branch/default'])
    assert exc_info.value.args == (b'unknown', )
    with pytest.raises(RevisionNotFound) as exc_info:
        resolve([sha1, b'unknown'])
    assert exc_info.value.args == (b'unknown', )
    with pytest.raises(RevisionNotFound) as exc_info:
        resolve([sha1, b'^unknown'])
    assert exc_info.value.args == (b'unknown', )

    # cases ignoring unknown revision
    assert resolve([b'unknown',
                    b'branch/default'
                    ],
                   ignore_unknown=True) == ({sha1}, empty)
    assert resolve([b'unknown',
                    b'^branch/default'
                    ],
                   ignore_unknown=True) == (empty, {sha1})
    assert resolve([b'^unknown',
                    b'branch/default'
                    ],
                   ignore_unknown=True) == ({sha1}, empty)
